"""Module for handling the Zoom service"""

import json
import os
import time
import typing
from gettext import gettext as _
from typing import Any, Dict, List, Tuple

from lutris import settings
from lutris.exceptions import AuthenticationError
from lutris.installer import AUTO_ELF_EXE, AUTO_WIN32_EXE
from lutris.installer.installer_file import InstallerFile
from lutris.installer.installer_file_collection import InstallerFileCollection
from lutris.runners import get_runner_human_name
from lutris.services.base import SERVICE_LOGIN, OnlineService
from lutris.services.service_game import ServiceGame
from lutris.services.service_media import ServiceMedia
from lutris.util import system
from lutris.util.http import HTTPError, Request
from lutris.util.log import logger
from lutris.util.strings import computer_size

if typing.TYPE_CHECKING:
    from lutris.installer.installer import LutrisInstaller


class ZoomCover(ServiceMedia):
    """Small size game logo"""

    service = "zoomplatform"
    size = (200, 300)
    dest_path = os.path.join(settings.CACHE_DIR, "zoom/cover/")
    file_patterns = ["%s.jpg"]
    api_field = "image"
    url_pattern = "%s"


class ZoomGame(ServiceGame):
    """Representation of a Zoom game"""

    service = "zoomplatform"

    @classmethod
    def new_from_zoom_game(cls, zoom_game):
        """Return a Zoom game instance from the API info"""
        service_game = ZoomGame()
        service_game.appid = str(zoom_game["id"])
        service_game.game_id = str(zoom_game["id"])
        service_game.slug = zoom_game["slug"]
        service_game.name = zoom_game["name"]
        details = zoom_game
        details["image"] = zoom_game["poster_url"]
        details["slug"] = zoom_game["slug"]
        service_game.details = json.dumps(details)
        return service_game


class ZoomService(OnlineService):
    """Service class for ZOOM Platform"""

    id = "zoomplatform"
    name = _("ZOOM Platform")
    icon = "zoom"
    has_extras = True
    drm_free = True
    medias = {"coverart": ZoomCover}
    default_format = "coverart"

    embed_url = "https://www.zoom-platform.com"
    api_url = "https://www.zoom-platform.com"

    redirect_uris = ["https://www.zoom-platform.com/account?li_token="]

    login_success_url = "https://www.zoom-platform.com/account?li_token="
    token_path = os.path.join(settings.CACHE_DIR, ".zoom.token")
    cookies_path = os.path.join(settings.CACHE_DIR, ".zoom.auth")
    cache_path = os.path.join(settings.CACHE_DIR, "zoom-library.json")

    runner_to_os_dict = {"wine": "windows", "linux": "linux"}

    @property
    def login_url(self):
        """Return authentication URL"""
        return f"{self.embed_url}/login?li=lutris&return_li_token=true"

    @property
    def credential_files(self) -> List[str]:
        return [self.cookies_path, self.token_path]

    def is_connected(self) -> bool:
        """Return whether the user is authenticated and if the service is available"""

        try:
            request = self.make_request(f"{self.api_url}/li/loggedin")
        except AuthenticationError:
            logger.debug("User is not authenticated")
            return False

        logger.debug("User is authenticated: %s", request)
        return True

    def is_authenticated(self):
        return self.is_connected()

    def load(self) -> List[ZoomGame]:
        """Load the user game library from the Zoom API"""
        if not self.is_connected():
            logger.error("User not connected to Zoom")
            return []
        games = [ZoomGame.new_from_zoom_game(game) for game in self.get_library()]
        for game in games:
            game.save()
        self.match_games()
        return games

    def login_callback(self, url) -> None:
        if "li_token=" not in url:
            logger.error("Login callback URL does not contain 'li_token'")
            return

        token = url.split("li_token=")[-1]
        with open(self.token_path, "w", encoding="utf-8") as token_file:
            token_file.write(token)

        assert not self.is_login_in_progress
        SERVICE_LOGIN.fire(self)

    def load_token(self) -> str:
        """Load token from disk"""
        if not os.path.exists(self.token_path):
            raise AuthenticationError("No Zoom token available")

        with open(self.token_path, encoding="utf-8") as token_file:
            token_content = token_file.read()

        if not token_content:
            raise AuthenticationError("No Zoom token available")

        return token_content

    def make_request(self, url: str) -> Any:
        """Send a token authenticated HTTP request to Zoom"""

        token = self.load_token()
        headers = {"Authorization": "Bearer " + token, "Accept": "application/json"}

        request = Request(url, headers=headers)
        request.get()
        return request.json

    def get_library(self) -> List[Dict]:
        """Return the user's library of Zoom games"""
        if system.path_exists(self.cache_path):
            logger.debug("Returning cached Zoom library")
            with open(self.cache_path, "r", encoding="utf-8") as zoom_cache:
                return json.load(zoom_cache)

        url = f"{self.api_url}/li/games"
        response = self.make_request(url)
        current_page = response.get("current_page", 0)
        total_pages = response.get("total_pages", 0)

        games = response["games"]
        while current_page < total_pages - 1:
            time.sleep(1)  # Avoid hitting the API too fast, until HTTP 429 is handled in the Request class
            logger.debug("Fetching additional pages of Zoom library")
            current_page += 1
            next_page_url = f"{url}?page={current_page}"
            response = self.make_request(next_page_url)
            games.extend(response["games"])

        # print(games)
        with open(self.cache_path, "w", encoding="utf-8") as zoom_cache:
            json.dump(games, zoom_cache)

        return games

    def get_extras(self, appid: str) -> Dict[str, List[dict]]:
        """Return a list of bonus content available for a Zoom ID and its DLCs"""
        logger.debug("Download extras for Zoom ID %s and its DLCs", appid)
        return {"extras": self._get_extra(appid)}

    def _get_extra(self, appid: str) -> List[dict]:
        # fetch the extra files urls and then parse the response to get the download url

        files_request = self.make_request(f"{self.api_url}/li/game/{appid}/files")
        # print(product_request)

        all_extras = []
        for extra_type in ["manual", "misc", "soundtrack"]:
            files = files_request[extra_type]
            for file in files:
                try:
                    download_request = self.make_request(f"{self.api_url}/li/download/{file['id']}")
                except HTTPError as ex:
                    logger.exception("Unable to locate extra '%s': %s", file.get("name"), ex)
                    continue

                extra_file_dict = {
                    "name": file["name"],
                    "url": download_request["url"],
                    "filename": file["name"],  # we cannot use "path" here as it can include a directory
                    "total_size": computer_size(file["size"]),
                }
                all_extras.append(extra_file_dict)
        return all_extras

    def generate_installer(self, db_game: Dict[str, Any]) -> Dict[str, Any]:
        logger.debug("Generating installer for %s", db_game)
        details = json.loads(db_game["details"])
        platforms = details["operating_systems"]
        if "linux" in platforms:
            return self._generate_installer("linux", db_game)
        else:
            return self._generate_installer("wine", db_game)

    def generate_installers(self, db_game: Dict[str, Any]) -> List[dict]:
        details = json.loads(db_game["details"])
        platforms = details["operating_systems"]

        installers = []

        if "linux" in platforms:
            installers.append(self._generate_installer("linux", db_game))

        if "windows" in platforms:
            installers.append(self._generate_installer("wine", db_game))

        if len(installers) > 1:
            for installer in installers:
                runner_human_name = get_runner_human_name(installer["runner"])
                installer["version"] += " " + (runner_human_name or installer["runner"])

        return installers

    def _generate_installer(self, runner: str, db_game: Dict[str, Any]) -> Dict[str, Any]:
        slug = db_game["slug"]
        system_config = {}
        if runner == "linux":
            game_config = {"exe": AUTO_ELF_EXE}
            script = [
                {"extract": {"file": "zoominstaller", "dst": "$CACHE"}},
                {"merge": {"src": "$CACHE", "dst": "$GAMEDIR"}},
            ]
        else:
            game_config = {"exe": AUTO_WIN32_EXE}
            script = [
                {"autosetup_zoom_platform": "zoominstaller"},
            ]
        return {
            "name": db_game["name"],
            "version": "Zoom",
            "slug": slug,
            "game_slug": self.get_installed_slug(db_game),
            "runner": runner,
            "zoomid": db_game["appid"],
            "script": {
                "game": game_config,
                "system": system_config,
                "files": [{"zoominstaller": "N/A:Select the installer from Zoom"}],
                "installer": script,
            },
        }

    def get_installer_files(
        self, installer: "LutrisInstaller", installer_file_id: str, selected_extras: List[dict]
    ) -> Tuple[List[InstallerFileCollection], List[InstallerFile]]:
        logger.debug("Getting installer files for %s", installer_file_id)
        platform = installer.runner
        if platform == "wine":
            platform = "windows"

        installer_files = self._get_installers(platform, installer.game_slug, installer.service_appid)
        files = [InstallerFileCollection(installer.game_slug, installer_file_id, installer_files)]
        extras = []

        if selected_extras:
            for selected_extra in selected_extras:
                extras.append(InstallerFile(installer.game_slug, selected_extra["name"], selected_extra))

        return files, extras

    def _get_installers(self, platform: str, game_slug: str, appid: str) -> List[InstallerFile]:
        # fetch the installer url and then parse the response to get the download url

        files_request = self.make_request(f"{self.api_url}/li/game/{appid}/files")
        # print(json)

        file_list = []
        files = files_request[platform]
        print(files)
        assert len(files) == 1, "More than one file found for %s" % platform  # TODO: Handle multiple files
        json = self.make_request(f"{self.api_url}/li/download/{files[0]['id']}")

        installer_file_dict = {
            "url": json["url"],
            "filename": files[0]["name"],
            "total_size": computer_size(files[0]["size"]),
        }

        installer_file = InstallerFile(
            game_slug,
            "zoominstaller",
            installer_file_dict,
        )
        file_list.append(installer_file)
        return file_list

    def get_service_game(self, zoom_game: dict) -> ZoomGame:
        return ZoomGame.new_from_zoom_game(zoom_game)

    def get_game_details(self, product_id: str) -> dict:
        logger.debug("Getting game details for %s", product_id)
        return {}

    def get_installed_runner_name(self, db_game: dict) -> str:
        platforms = [platform.casefold() for platform in self.get_game_platforms(db_game)]
        return "linux" if "linux" in platforms else "wine"



"""Xenia Runner (Wine-based)"""

import os
from gettext import gettext as _

from lutris import settings
from lutris.exceptions import MissingExecutableError, MissingGameExecutableError
from lutris.runners.wine import wine
from lutris.util import system
from lutris.util.wine.wine import get_default_wine_version


class xenia(wine):
    human_name = _("Xenia")
    description = _("Xbox 360 Emulator")
    platforms = [_("Microsoft Xbox 360")]
    runnable_alone = True
    multiple_versions = False
    runner_executable = "xenia/xenia_canary.exe"
    download_url = (
        "https://github.com/xenia-canary/xenia-canary-releases/releases/latest/download/xenia_canary_windows.zip"
    )
    entry_point_option = "main_file"

    game_options = [
        {
            "option": "main_file",
            "type": "file",
            "label": _("Game file"),
            "help": _("Path to .xex or .iso file"),
            "default_path": "game_path",
        },
        {
            "option": "prefix",
            "type": "directory",
            "label": _("Wine prefix"),
            "help": _("Wine prefix for Xenia. Leave empty for default."),
        },
        {
            "option": "arch",
            "type": "choice",
            "label": _("Prefix architecture"),
            "choices": [(_("64-bit"), "win64")],
            "default": "win64",
            "help": _("Xenia requires a 64-bit Wine prefix"),
        },
    ]

    runner_options = [
        {
            "option": "fullscreen",
            "type": "bool",
            "label": _("Fullscreen"),
            "default": True,
        },
    ] + wine.runner_options

    system_options_override = [{"option": "disable_runtime", "default": True}]

    @property
    def game_exe(self):
        """Return path to the managed Xenia Windows binary."""
        return os.path.join(settings.RUNNER_DIR, self.runner_executable)

    @property
    def prefix_path(self):
        """Return prefix path, defaulting to a location in the runner directory."""
        prefix = self.game_config.get("prefix")
        if prefix:
            return os.path.expanduser(prefix)
        return os.path.join(self.directory, "prefix")

    @property
    def wine_arch(self):
        """Xenia requires 64-bit."""
        return "win64"

    def read_version_from_config(self, default=None):
        """Read Wine version from config using the correct runner slug."""
        for level in [self.config.game_level, self.config.runner_level]:
            if self.name in level:
                runner_version = level[self.name].get("version")
                if runner_version:
                    return runner_version
        if default:
            return default
        return get_default_wine_version()

    def is_installed(self, flatpak_allowed=True, version=None, fallback=True):
        """Check if the Xenia binary is installed."""
        return os.path.isfile(self.game_exe)

    def play(self):
        """Launch an Xbox 360 game through Xenia under Wine."""
        launch_info = {"env": self.get_env(os_env=False)}

        xenia_exe = self.game_exe
        if not system.path_exists(xenia_exe):
            raise MissingExecutableError(_("Xenia executable not found at '%s'") % xenia_exe)

        command = self.get_command()
        command.append(xenia_exe)

        if self.runner_config.get("fullscreen"):
            command.append("--fullscreen")

        game_path = self.game_config.get("main_file") or ""
        if not system.path_exists(game_path):
            raise MissingGameExecutableError(filename=game_path)
        command.append(game_path)

        launch_info["command"] = command
        return launch_info



"""Reads the Dolphin game database, stored in a binary format"""

import os

from lutris.util.log import logger

DOLPHIN_GAME_CACHE_FILE = os.path.expanduser("~/.cache/dolphin-emu/gamelist.cache")
SUPPORTED_CACHE_VERSION = 24


def get_hex_string(string):
    """Return the hexadecimal representation of a string"""
    return " ".join("{:02x}".format(c) for c in string)


def get_word_len(string):
    """Return the length of a string as specified in the Dolphin format"""
    return int("0x" + "".join("{:02x}".format(c) for c in string[::-1]), 0)


# https://github.com/dolphin-emu/dolphin/blob/90a994f93780ef8a7cccfc02e00576692e0f2839/Source/Core/UICommon/GameFile.h#L140
# https://github.com/dolphin-emu/dolphin/blob/90a994f93780ef8a7cccfc02e00576692e0f2839/Source/Core/UICommon/GameFile.cpp#L318


class DolphinCacheReader:
    header_size = 20
    structure = {
        "valid": "b",
        "file_path": "s",
        "file_name": "s",
        "file_size": 8,
        "volume_size": 8,
        "volume_size_type": 4,
        "is_datel_disc": "b",
        "is_nkit": "b",
        "short_names": "a",
        "long_names": "a",
        "short_makers": "a",
        "long_makers": "a",
        "descriptions": "a",
        "internal_name": "s",
        "game_id": "s",
        "gametdb_id": "s",
        "title_id": 8,
        "maker_id": "s",
        "region": 4,
        "country": 4,
        "platform": 1,
        "platform_": 3,
        "blob_type": 4,
        "block_size": 8,
        "compression_method": "s",
        "revision": 2,
        "disc_number": 1,
        "apploader_date": "s",
        "custom_name": "s",
        "custom_description": "s",
        "custom_maker": "s",
        "volume_banner": "i",
        "custom_banner": "i",
        "default_cover": "c",
        "custom_cover": "c",
    }

    def __init__(self):
        self.offset = 0
        with open(DOLPHIN_GAME_CACHE_FILE, "rb") as dolphin_cache_file:
            self.cache_content = dolphin_cache_file.read()
        cache_version = get_word_len(self.cache_content[:4])
        if cache_version != SUPPORTED_CACHE_VERSION:
            logger.warning("Dolphin cache version expected %s but found %s", SUPPORTED_CACHE_VERSION, cache_version)

    def get_game(self):
        game = {}
        for key, i in self.structure.items():
            if i == "s":
                game[key] = self.get_string()
            elif i == "b":
                game[key] = self.get_boolean()
            elif i == "a":
                game[key] = self.get_array()
            elif i == "i":
                game[key] = self.get_image()
            elif i == "c":
                game[key] = self.get_cover()
            else:
                game[key] = self.get_raw(i)
        return game

    def get_games(self):
        self.offset += self.header_size
        games = []
        while self.offset < len(self.cache_content):
            try:
                games.append(self.get_game())
            except Exception as ex:
                logger.error("Failed to read Dolphin database: %s", ex)
        return games

    def get_boolean(self):
        res = bool(get_word_len(self.cache_content[self.offset : self.offset + 1]))
        self.offset += 1
        return res

    def get_array(self):
        array_len = get_word_len(self.cache_content[self.offset : self.offset + 4])
        self.offset += 4
        array = {}
        for _i in range(array_len):
            array_key = self.get_raw(4)
            array[array_key] = self.get_string()
        return array

    def get_image(self):
        data_len = get_word_len(self.cache_content[self.offset : self.offset + 4])
        self.offset += 4
        res = self.cache_content[self.offset : self.offset + data_len * 4]  # vector<u32>
        self.offset += data_len * 4
        width = get_word_len(self.cache_content[self.offset : self.offset + 4])
        self.offset += 4
        height = get_word_len(self.cache_content[self.offset : self.offset + 4])
        self.offset += 4
        return (width, height), res

    def get_cover(self):
        array_len = get_word_len(self.cache_content[self.offset : self.offset + 4])
        self.offset += 4
        return self.get_raw(array_len)

    def get_raw(self, word_len):
        res = get_hex_string(self.cache_content[self.offset : self.offset + word_len])
        self.offset += word_len
        return res

    def get_string(self):
        word_len = get_word_len(self.cache_content[self.offset : self.offset + 4])
        self.offset += 4
        string = self.cache_content[self.offset : self.offset + word_len]
        self.offset += word_len
        return string.decode("utf8")



import os

from lutris.util.system import get_md5_hash


def get_folder_contents(target_directory: str, with_hash: bool = True) -> list:
    """Recursively iterate over a folder content and return its details"""
    folder_content = []
    for path, dir_names, file_names in os.walk(target_directory):
        for dir_name in dir_names:
            dir_path = os.path.join(path, dir_name)
            folder_content.append(
                {
                    "name": dir_path,
                    "date_created": int(os.path.getctime(dir_path)),
                    "date_modified": int(os.path.getmtime(dir_path)),
                    "date_accessed": int(os.path.getatime(dir_path)),
                    "type": "folder",
                }
            )
        for file_name in file_names:
            file_path = os.path.join(path, file_name)
            file_stats = os.stat(file_path)
            file_desc = {
                "name": file_path,
                "size": file_stats.st_size,
                "date_created": int(file_stats.st_ctime),
                "date_modified": int(file_stats.st_mtime),
                "date_accessed": int(file_stats.st_atime),
                "type": "file",
            }
            if with_hash:
                file_desc["md5_hash"] = get_md5_hash(file_path)
            folder_content.append(file_desc)
    return folder_content



# Standard Library
from gettext import gettext as _

from lutris.exceptions import MissingGameExecutableError

# Lutris Modules
from lutris.runners.runner import Runner
from lutris.util import system


class osmose(Runner):
    human_name = _("Osmose")
    description = _("Sega Master System Emulator")
    platforms = [_("Sega Master System")]
    runner_executable = "osmose/osmose"
    game_options = [
        {
            "option": "main_file",
            "type": "file",
            "label": _("ROM file"),
            "default_path": "game_path",
            "help": _(
                "The game data, commonly called a ROM image.\n"
                "Supported formats: SMS and GG files. ZIP compressed "
                "ROMs are supported."
            ),
        }
    ]
    runner_options = [
        {
            "option": "fullscreen",
            "type": "bool",
            "label": _("Fullscreen"),
            "default": False,
        }
    ]

    def play(self):
        """Run Sega Master System game"""
        arguments = self.get_command()
        rom = self.game_config.get("main_file") or ""
        if not system.path_exists(rom):
            raise MissingGameExecutableError(filename=rom)
        arguments.append(rom)
        if self.runner_config.get("fullscreen"):
            arguments.append("-fs")
            arguments.append("-bilinear")
        return {"command": arguments}



import configparser
import os
from typing import Any

from lutris.gui.widgets import NotificationSource
from lutris.util.log import logger


class SettingsIO:
    """ConfigParser abstraction."""

    def __init__(self, config_file: str):
        self.config_file = config_file
        self.config = configparser.ConfigParser()

        # A notification that fires on each settings change
        self.SETTINGS_CHANGED = NotificationSource()  # called with (setting-key, new-value, section)

        if os.path.exists(self.config_file):
            try:
                self.config.read([self.config_file])
            except configparser.ParsingError as ex:
                logger.error("Failed to readconfig file %s: %s", self.config_file, ex)
            except UnicodeDecodeError as ex:
                logger.error("Some invalid characters are preventing the setting file from loading properly: %s", ex)

    def read_setting(self, key: str, default: str = "", section: str = "lutris") -> str:
        """Read a setting from the config file

        Params:
            key (str): Setting key
            section (str): Optional section, default to 'lutris'
            default (str): Default value to return if setting not present
        """
        try:
            return self.config.get(section, key)
        except (configparser.NoOptionError, configparser.NoSectionError):
            return default

    def read_bool_setting(self, key: str, default: bool = False, section: str = "lutris") -> bool:
        text = self.read_setting(key, "", section=section).casefold()
        if text == "true":
            return True
        if text == "false":
            return False

        return default

    def write_setting(self, key: str, value: Any, section: str = "lutris") -> None:
        if not self.config.has_section(section):
            self.config.add_section(section)
        self.config.set(section, key, str(value))

        with open(self.config_file, "w", encoding="utf-8") as config_file:
            self.config.write(config_file)

        self.SETTINGS_CHANGED.fire(key, value, section)



"""Nvidia library detection from Proton"""

import os
from ctypes import CDLL, POINTER, Structure, addressof, c_char_p, c_int, c_void_p, cast

from lutris.util.log import logger

RTLD_DI_LINKMAP = 2


class LinkMap(Structure):
    """
    from dlinfo(3)

    struct link_map {
        ElfW(Addr) l_addr;  /* Difference between the
                               address in the ELF file and
                               the address in memory */
        char      *l_name;  /* Absolute pathname where
                               object was found */
        ElfW(Dyn) *l_ld;    /* Dynamic section of the
                               shared object */
        struct link_map *l_next, *l_prev;
                            /* Chain of loaded objects */
        /* Plus additional fields private to the implementation */
    };
    """

    _fields_ = [("l_addr", c_void_p), ("l_name", c_char_p), ("l_ld", c_void_p)]


def get_nvidia_glx_path():
    """Return the absolute path to the libGLX_nvidia library"""
    try:
        libdl = CDLL("libdl.so.2")
    except OSError:
        logger.error("Unable to load libdl.so.2")
        return None

    try:
        libglx_nvidia = CDLL("libGLX_nvidia.so.0")
    except OSError:
        return None

    # from dlinfo(3)
    #
    # int dlinfo (void *restrict handle, int request, void *restrict info)
    dlinfo_func = libdl.dlinfo
    dlinfo_func.argtypes = c_void_p, c_int, c_void_p
    dlinfo_func.restype = c_int

    # Allocate a LinkMap object
    glx_nvidia_info_ptr = POINTER(LinkMap)()

    # Run dlinfo(3) on the handle to libGLX_nvidia.so.0, storing results at the
    # address represented by glx_nvidia_info_ptr
    if dlinfo_func(libglx_nvidia._handle, RTLD_DI_LINKMAP, addressof(glx_nvidia_info_ptr)) != 0:
        logger.error("Unable to read Nvidia information")
        return None

    # Grab the contents our of our pointer
    glx_nvidia_info = cast(glx_nvidia_info_ptr, POINTER(LinkMap)).contents

    # Decode the path to our library to a str()
    if glx_nvidia_info.l_name is None:
        logger.error("Error reading the Nvidia library path")
        return None
    try:
        libglx_nvidia_path = os.fsdecode(glx_nvidia_info.l_name)
    except UnicodeDecodeError as ex:
        logger.error("Error decoding the Nvidia library path: %s", ex)
        return None

    # Follow any symlinks to the actual file
    return os.path.realpath(libglx_nvidia_path)


def get_nvidia_dll_path():
    """Return the path to the location of DLL files for use by Wine/Proton
    from the NVIDIA Linux driver.
    See https://gitlab.steamos.cloud/steamrt/steam-runtime-tools/-/issues/71 for
    background on the chosen method of DLL discovery.
    """
    libglx_path = get_nvidia_glx_path()
    if not libglx_path:
        return
    nvidia_wine_dir = os.path.join(os.path.dirname(libglx_path), "nvidia/wine")
    if os.path.exists(os.path.join(nvidia_wine_dir, "nvngx.dll")):
        return nvidia_wine_dir



import struct
from io import BytesIO

# for the record:
#  - int32 = signed varint
#  - int64 = signed varint
#  - enum = signed varint
#  - uint32 = unsigned varint
#  - uint64 = unsigned varint
#  - sint32 = zigzag signed varint
#  - sint64 = zigzag signed varint

# https://developers.google.com/protocol-buffers/docs/encoding
# 0  Varint             int32, int64, uint32, uint64, sint32, sint64, bool, enum
# 1  64-bit             fixed64, sfixed64, double
# 2  Length-delimited   string, bytes, embedded messages, packed repeated fields
# 3  Start group        groups (deprecated)
# 4  End group	        groups (deprecated)
# 5  32-bit             fixed32, sfixed32, float


class PrimativeType:
    @staticmethod
    def decode(data):
        raise NotImplementedError()


class type_double(PrimativeType):
    wire_type = 1

    @staticmethod
    def decode(data):
        # data = 64-bit
        (val,) = struct.unpack("<d", data)
        return val


class type_float(PrimativeType):
    wire_type = 5

    @staticmethod
    def decode(data):
        # data = 32-bit
        (val,) = struct.unpack("<f", data)
        return val


class type_int32(PrimativeType):
    wire_type = 0

    @staticmethod
    def decode(data):
        # data = signed varint
        val = Message.signed_to_long(data, 32)
        return int(val)


class type_int64(PrimativeType):
    wire_type = 0

    @staticmethod
    def decode(data):
        # data = signed varint
        val = Message.signed_to_long(data, 64)
        return int(val)


class type_uint32(PrimativeType):
    wire_type = 0

    @staticmethod
    def decode(data):
        # data = unsigned varint
        return int(data)


class type_uint64(PrimativeType):
    wire_type = 0

    @staticmethod
    def decode(data):
        # data = unsigned varint
        return int(data)


class type_sint32(PrimativeType):
    wire_type = 0

    @staticmethod
    def decode(data):
        # data = zigzag signed varint
        val = Message.signed_to_long(data, 32)
        val = Message.zigzag_to_long(val)
        return int(val)


class type_sint64(PrimativeType):
    wire_type = 0

    @staticmethod
    def decode(data):
        # data = zigzag signed varint
        val = Message.signed_to_long(data, 64)
        val = Message.zigzag_to_long(val)
        return int(val)


class type_fixed32(PrimativeType):
    wire_type = 5

    @staticmethod
    def decode(data):
        # data = 32-bit
        (val,) = struct.unpack("<I", data)
        return int(val)


class type_fixed64(PrimativeType):
    wire_type = 1

    @staticmethod
    def decode(data):
        # data = 64-bit
        (val,) = struct.unpack("<Q", data)
        return int(val)


class type_sfixed32(PrimativeType):
    wire_type = 5

    @staticmethod
    def decode(data):
        # data = 32-bit
        (val,) = struct.unpack("<i", data)
        return int(val)


class type_sfixed64(PrimativeType):
    wire_type = 1

    @staticmethod
    def decode(data):
        # data = 64-bit
        (val,) = struct.unpack("<q", data)
        return int(val)


class type_bool(PrimativeType):
    wire_type = 0

    @staticmethod
    def decode(data):
        # data = signed varint
        return data != 0


class type_string(PrimativeType):
    wire_type = 2

    @staticmethod
    def decode(data):
        # data = binary string
        return data


class type_bytes(PrimativeType):
    wire_type = 2

    @staticmethod
    def decode(data):
        # data = binary string
        return data


# probs best to go with int64
type_enum = type_int64


class Message:
    @staticmethod
    def read_varint(stream):
        res = 0
        i = 0
        while 1:
            c = ord(stream.read(1))
            res |= (c & 127) << (i * 7)
            if c & 128 == 0:
                break
            i += 1
        return res

    @staticmethod
    def signed_to_long(x, bits):
        "converts a previously read signed varint into a long"
        if x > 0x7FFFFFFFFFFFFFFF:
            x -= 1 << 64
            x |= ~((1 << bits) - 1)
        else:
            x &= (1 << bits) - 1
        return x

    # zigzag conversion from google
    # https://github.com/google/protobuf/blob/master/python/google/protobuf/internal/wire_format.py

    @staticmethod
    def zigzag_to_long(x):
        if not x & 0x1:
            return x >> 1
        return (x >> 1) ^ (~0)

    @staticmethod
    def read_tag(stream):
        var = Message.read_varint(stream)
        field_number = var >> 3
        wire_type = var & 7

        if wire_type == 0:
            data = Message.read_varint(stream)
        elif wire_type == 1:
            data = stream.read(8)
        elif wire_type == 2:
            length = Message.read_varint(stream)
            data = stream.read(length)
        elif wire_type in (3, 4):
            raise NotImplementedError("groups are deprecated")
        elif wire_type == 5:
            data = stream.read(4)
        else:
            raise TypeError("unknown wire type (%d)" % wire_type)
        return (field_number, wire_type, data)

    def lookup_id(self, _id):
        for _, i in enumerate(self.__lookup__):
            if i[3] == _id:
                return i

    def decode(self, s: bytes):
        f = BytesIO(s)
        while f.tell() < len(s):
            field_number, _, data = self.read_tag(f)
            field = self.lookup_id(field_number)
            if not field:
                continue

            field_multiplicity, field_type, field_name, _ = field
            if issubclass(field_type, PrimativeType):
                value = field_type.decode(data)
            elif issubclass(field_type, Message):
                value = field_type()
                value.decode(data)
            else:
                raise TypeError("field type must be a subclass of PrimativeType or Message")

            if field_multiplicity == "repeated":
                if getattr(self, field_name) is None:
                    # if not isinstance(getattr(self, field_name), list):
                    # ? what if the attribute was already filled with data ?
                    setattr(self, field_name, [])
                getattr(self, field_name).append(value)
            else:
                setattr(self, field_name, value)

    def __lookup__(self):
        return



"""Epic Games Store service"""

import json
import os
from gettext import gettext as _
from typing import Any, Dict, Optional

import requests
from gi.repository import Gio

from lutris import settings
from lutris.config import LutrisConfig, write_game_config
from lutris.database.games import add_game, get_game_by_field
from lutris.database.services import ServiceGameCollection
from lutris.game import Game
from lutris.gui.widgets.utils import Image, paste_overlay, thumbnail_image
from lutris.services.base import SERVICE_LOGIN, AuthTokenExpiredError, OnlineService
from lutris.services.lutris import sync_media
from lutris.services.service_game import ServiceGame
from lutris.services.service_media import ServiceMedia
from lutris.util import system
from lutris.util.egs.egs_launcher import EGSLauncher
from lutris.util.log import logger
from lutris.util.strings import slugify

EGS_GAME_ART_PATH = os.path.expanduser("~/.cache/lutris/egs/game_box")
EGS_GAME_BOX_PATH = os.path.expanduser("~/.cache/lutris/egs/game_box_tall")
EGS_LOGO_PATH = os.path.expanduser("~/.cache/lutris/egs/game_logo")
EGS_BANNERS_PATH = os.path.expanduser("~/.cache/lutris/egs/banners")
EGS_BOX_ART_PATH = os.path.expanduser("~/.cache/lutris/egs/boxart")
BANNER_SIZE = (316, 178)
BOX_ART_SIZE = (200, 267)


class DieselGameMedia(ServiceMedia):
    service = "egs"
    remote_size = (200, 267)
    file_patterns = ["%s.jpg"]
    min_logo_x = 300
    min_logo_y = 150

    def _render_filename(self, filename):
        game_box_path = os.path.join(self.dest_path, filename)
        logo_path = os.path.join(EGS_LOGO_PATH, filename.replace(".jpg", ".png"))
        has_logo = os.path.exists(logo_path)
        thumb_image = Image.open(game_box_path)
        thumb_image = thumb_image.convert("RGBA")
        thumb_image = thumbnail_image(thumb_image, self.remote_size)
        if has_logo:
            logo_image = Image.open(logo_path)
            logo_image = logo_image.convert("RGBA")
            logo_width, logo_height = logo_image.size
            if logo_width > self.min_logo_x:
                logo_image = logo_image.resize(
                    (self.min_logo_x, int(logo_height * (self.min_logo_x / logo_width))),
                    resample=Image.Resampling.BICUBIC,
                )
            elif logo_height > self.min_logo_y:
                logo_image = logo_image.resize(
                    (int(logo_width * (self.min_logo_y / logo_height)), self.min_logo_y),
                    resample=Image.Resampling.BICUBIC,
                )
            thumb_image = paste_overlay(thumb_image, logo_image)
        thumb_path = os.path.join(self.dest_path, filename)
        thumb_image = thumb_image.convert("RGB")
        thumb_image.save(thumb_path)

    def get_media_url(self, details: Dict[str, Any]) -> Optional[str]:
        for image in details.get("keyImages", []):
            if image["type"] == self.api_field:
                return image["url"] + "?w=%s&resize=1&h=%s" % (self.remote_size[0], self.remote_size[1])
        return None


class DieselGameBoxTall(DieselGameMedia):
    """EGS tall game box"""

    size = (200, 267)
    remote_size = size
    min_logo_x = 100
    min_logo_y = 100
    dest_path = os.path.join(settings.CACHE_DIR, "egs/game_box_tall")
    api_field = "DieselGameBoxTall"

    def render(self):
        for filename in os.listdir(self.dest_path):
            self._render_filename(filename)


class DieselGameBoxSmall(DieselGameBoxTall):
    size = (100, 133)
    remote_size = (200, 267)


class DieselGameBox(DieselGameBoxTall):
    """EGS game box"""

    size = (316, 178)
    remote_size = size
    min_logo_x = 300
    min_logo_y = 150
    dest_path = os.path.join(settings.CACHE_DIR, "egs/game_box")
    api_field = "DieselGameBox"


class DieselGameBannerSmall(DieselGameBox):
    size = (158, 89)
    remote_size = (316, 178)


class DieselGameBoxLogo(DieselGameMedia):
    """EGS game box"""

    size = (200, 100)
    remote_size = size
    file_patterns = ["%s.png"]
    visible = False
    dest_path = os.path.join(settings.CACHE_DIR, "egs/game_logo")
    api_field = "DieselGameBoxLogo"


class EGSGame(ServiceGame):
    """Service game for Epic Games Store"""

    service = "egs"

    @classmethod
    def new_from_api(cls, egs_game):
        """Convert an EGS game to a service game"""
        service_game = cls()
        service_game.appid = egs_game["appName"]
        service_game.slug = slugify(egs_game["title"])
        service_game.name = egs_game["title"]
        service_game.details = json.dumps(egs_game)
        return service_game


class EpicGamesStoreService(OnlineService):
    """Service class for Epic Games Store"""

    id = "egs"
    name = _("Epic Games Store")
    login_window_width = 500
    login_window_height = 850
    icon = "egs"
    online = True
    runner = "wine"
    client_installer = "epic-games-store"
    medias = {
        "game_box_small": DieselGameBoxSmall,
        "game_banner_small": DieselGameBannerSmall,
        "game_box": DieselGameBox,
        "box_tall": DieselGameBoxTall,
    }
    extra_medias = {
        "logo": DieselGameBoxLogo,
    }
    default_format = "game_banner_small"
    requires_login_page = True
    cookies_path = os.path.join(settings.CACHE_DIR, ".egs.auth")
    token_path = os.path.join(settings.CACHE_DIR, ".egs.token")
    cache_path = os.path.join(settings.CACHE_DIR, "egs-library.json")
    login_url = (
        "https://www.epicgames.com/id/login?redirectUrl="
        "https%3A//www.epicgames.com/id/api/redirect%3F"
        "clientId%3D34a02cf8f4414e29b15921876da36f9a%26responseType%3Dcode"
    )
    redirect_uris = ["https://www.epicgames.com/id/api/redirect"]
    oauth_url = "https://account-public-service-prod03.ol.epicgames.com"
    catalog_url = "https://catalog-public-service-prod06.ol.epicgames.com"
    library_url = "https://library-service.live.use1a.on.epicgames.com"

    user_agent = (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
        "AppleWebKit/537.36 (KHTML, like Gecko) "
        "EpicGamesLauncher/11.0.1-14907503+++Portal+Release-Live "
        "UnrealEngine/4.23.0-14907503+++Portal+Release-Live "
        "Chrome/84.0.4147.38 Safari/537.36"
    )

    def __init__(self):
        super().__init__()
        self.session = requests.session()
        self.session.headers["User-Agent"] = self.user_agent
        self.session_data = {}
        if os.path.exists(self.token_path):
            try:
                with open(self.token_path, encoding="utf-8") as token_file:
                    self.session_data = json.loads(token_file.read())
            except Exception as ex:
                logger.exception("Unable to load token file '%s': %s", token_file, ex)

    @property
    def http_basic_auth(self):
        return requests.auth.HTTPBasicAuth("34a02cf8f4414e29b15921876da36f9a", "daafbccc737745039dffe53d94fc76cf")

    def is_connected(self):
        return self.is_authenticated()

    def login_callback(self, content):
        """Once the user logs in in a browser window, Epic redirects
        to a page containing a Session ID which we can use to finish the authentication.
        Store session ID and exchange token to auth file"""
        logger.debug("Login to EGS successful")
        logger.debug(content)
        content_json = json.loads(content.decode())
        session_id = content_json["authorizationCode"]

        self.start_session(authorization_code=session_id)
        SERVICE_LOGIN.fire(self)

    def resume_session(self):
        self.session.headers["Authorization"] = "bearer %s" % self.session_data["access_token"]
        response = self.session.get("%s/account/api/oauth/verify" % self.oauth_url)
        if response.status_code >= 500:
            response.raise_for_status()

        response_content = response.json()
        if "errorMessage" in response_content:
            raise RuntimeError(response_content)
        return response_content

    def start_session(self, exchange_code=None, authorization_code=None):
        if exchange_code:
            params = {"grant_type": "exchange_code", "exchange_code": exchange_code, "token_type": "eg1"}
        elif authorization_code:
            params = {"grant_type": "authorization_code", "code": authorization_code, "token_type": "eg1"}
        else:
            params = {
                "grant_type": "refresh_token",
                "refresh_token": self.session_data["refresh_token"],
                "token_type": "eg1",
            }

        response = self.session.post(
            "https://account-public-service-prod03.ol.epicgames.com/account/api/oauth/token",
            data=params,
            auth=self.http_basic_auth,
        )
        if response.status_code >= 500:
            response.raise_for_status()

        response_content = response.json()
        if "error" in response_content:
            raise RuntimeError(response_content)
        with open(self.token_path, "w", encoding="utf-8") as auth_file:
            auth_file.write(json.dumps(response_content, indent=2))
        self.session_data = response_content

    def get_game_details(self, asset):
        namespace = asset["namespace"]
        catalog_item_id = asset["catalogItemId"]
        response = self.session.get(
            "%s/catalog/api/shared/namespace/%s/bulk/items" % (self.catalog_url, namespace),
            params={
                "id": catalog_item_id,
                "includeDLCDetails": True,
                "includeMainGameDetails": True,
                "country": "US",
                "locale": "en",
            },
        )
        response.raise_for_status()
        # Merge the details with the initial asset to keep 'appName'
        asset.update(response.json()[catalog_item_id])
        return asset

    def get_library(self):
        self.resume_session()
        response = self.session.get(
            "%s/library/api/public/items" % self.library_url, params={"includeMetadata": "true"}
        )
        response.raise_for_status()
        resData = response.json()
        records = resData["records"]
        cursor = resData["responseMetadata"].get("nextCursor", None)
        while cursor:
            response = self.session.get(
                "%s/library/api/public/items" % self.library_url, params={"includeMetadata": "true", "cursor": cursor}
            )
            response.raise_for_status()
            resData = response.json()
            records.extend(resData["records"])
            cursor = resData["responseMetadata"].get("nextCursor", None)

        games = []
        for record in records:
            if record["namespace"] == "ue":
                continue
            game_details = self.get_game_details(record)
            games.append(game_details)
        return games

    def load(self):
        """Load the list of games"""
        try:
            library = self.get_library()
        except Exception as ex:  # pylint=disable:broad-except
            logger.warning("EGS Token expired")
            raise AuthTokenExpiredError("EGS Token expired") from ex
        egs_games = []
        for game in library:
            try:
                # Subscriptions turn up as 'games' that have no 'appName'; these
                # are not really games, so we skip them.
                if "appName" in game:
                    egs_game = EGSGame.new_from_api(game)
                else:
                    continue
            except Exception as ex:
                logger.exception("Unable to interpret EGS game: %s", ex)
                logger.info("EGS game skipped: %s", game)
                continue
            egs_game.save()
            egs_games.append(egs_game)
        return egs_games

    def install_from_egs(self, egs_game, manifest):
        """Create a new Lutris game based on an existing EGS install"""
        app_name = manifest["AppName"]
        logger.debug("Installing EGS game %s", app_name)
        service_game = ServiceGameCollection.get_game("egs", app_name)
        if not service_game:
            logger.error("Aborting install, %s is not present in the game library.", app_name)
            return
        lutris_game_id = slugify(service_game["name"]) + "-" + self.id
        existing_game = get_game_by_field(lutris_game_id, "installer_slug")
        if existing_game:
            return
        game_config = LutrisConfig(game_config_id=egs_game["configpath"]).game_level
        game_config["game"]["args"] = get_launch_arguments(app_name)
        configpath = write_game_config(lutris_game_id, game_config)
        slug = self.get_installed_slug(service_game)
        add_game(
            name=service_game["name"],
            runner=egs_game["runner"],
            slug=slug,
            directory=egs_game["directory"],
            installed=1,
            installer_slug=lutris_game_id,
            configpath=configpath,
            service=self.id,
            service_id=app_name,
        )
        return slug

    def add_installed_games(self):
        """Scan an existing EGS install for games"""
        egs_game = get_game_by_field("epic-games-store", "slug")
        if not egs_game:
            logger.error("EGS is not installed in Lutris")
            return

        egs_prefix = egs_game["directory"].split("drive_c")[0]
        logger.info("EGS detected in %s", egs_prefix)
        if not system.path_exists(os.path.join(egs_prefix, "drive_c")):
            logger.error("Invalid install of EGS at %s", egs_prefix)
            return
        egs_launcher = EGSLauncher(egs_prefix)
        installed_slugs = []
        for manifest in egs_launcher.iter_manifests():
            slug = self.install_from_egs(egs_game, manifest)
            if slug:
                installed_slugs.append(slug)
        sync_media(installed_slugs)
        logger.debug("All EGS games imported")

    def generate_installer(self, db_game, egs_db_game):
        egs_game = Game(egs_db_game["id"])
        egs_exe = egs_game.config.game_config["exe"]
        if not os.path.isabs(egs_exe):
            egs_exe = os.path.join(egs_game.config.game_config["prefix"], egs_exe)
        return {
            "name": db_game["name"],
            "version": self.name,
            "slug": slugify(db_game["name"]) + "-" + self.id,
            "game_slug": self.get_installed_slug(db_game),
            "runner": self.get_installed_runner_name(db_game),
            "appid": db_game["appid"],
            "script": {
                "requires": self.client_installer,
                "game": {
                    "args": get_launch_arguments(db_game["appid"]),
                },
                "installer": [
                    {
                        "task": {
                            "name": "wineexec",
                            "executable": egs_exe,
                            "args": get_launch_arguments(db_game["appid"], "install"),
                            "prefix": egs_game.config.game_config["prefix"],
                            "description": (
                                "The Epic Game Store will now open. Please launch "
                                "the installation of %s then close the EGS client "
                                "once the game has been downloaded." % db_game["name"]
                            ),
                        }
                    }
                ],
            },
        }

    def get_installed_runner_name(self, db_game):
        return self.runner

    def install(self, db_game):
        egs_game = get_game_by_field(self.client_installer, "slug")
        application = Gio.Application.get_default()
        if not egs_game or not egs_game["installed"]:
            logger.warning("EGS (%s) not installed", self.client_installer)
            application.show_lutris_installer_window(game_slug=self.client_installer)
        else:
            application.show_installer_window(
                [self.generate_installer(db_game, egs_game)], service=self, appid=db_game["appid"]
            )


def get_launch_arguments(app_name, action="launch"):
    return ("-opengl -SkipBuildPatchPrereq -com.epicgames.launcher://apps/%s?action=%s") % (app_name, action)



"""Internal settings."""

import json
import os
import sys
from gettext import gettext as _
from typing import Dict, cast

from gi.repository import GLib

from lutris import __version__
from lutris.util.log import logger
from lutris.util.settings import SettingsIO

PROJECT = "Lutris"
VERSION = __version__
COPYRIGHT = _("(c) 2009 Lutris Team")
AUTHORS = [_("The Lutris team")]

# Paths
CONFIG_DIR = os.path.join(GLib.get_user_config_dir(), "lutris")
DATA_DIR = os.path.join(GLib.get_user_data_dir(), "lutris")
if not os.path.exists(CONFIG_DIR):
    # Set the config dir to ~/.local/share/lutris as we're deprecating ~/.config/lutris
    CONFIG_DIR = DATA_DIR
CONFIG_FILE = os.path.join(CONFIG_DIR, "lutris.conf")
sio = SettingsIO(CONFIG_FILE)

RUNNER_DIR = sio.read_setting("runner_dir") or os.path.join(DATA_DIR, "runners")
RUNTIME_DIR = sio.read_setting("runtime_dir") or os.path.join(DATA_DIR, "runtime")
CACHE_DIR = sio.read_setting("cache_dir") or os.path.join(GLib.get_user_cache_dir(), "lutris")
TMP_DIR = os.path.join(CACHE_DIR, "tmp")
GAME_CONFIG_DIR = os.path.join(CONFIG_DIR, "games")
RUNNERS_CONFIG_DIR = os.path.join(CONFIG_DIR, "runners")
WINE_DIR: str = os.path.join(RUNNER_DIR, "wine")

SHADER_CACHE_DIR = os.path.join(CACHE_DIR, "shaders")
INSTALLER_CACHE_DIR = os.path.join(CACHE_DIR, "installer")
BANNER_PATH = os.path.join(DATA_DIR, "banners")
COVERART_PATH = os.path.join(DATA_DIR, "coverart")

RUNTIME_VERSIONS_PATH = os.path.join(CACHE_DIR, "versions.json")
ICON_PATH = os.path.join(GLib.get_user_data_dir(), "icons", "hicolor", "128x128", "apps")

if "nosetests" in sys.argv[0] or "nose2" in sys.argv[0] or "pytest" in sys.argv[0]:
    DB_PATH = "/tmp/pga.db"
else:
    DB_PATH = sio.read_setting("pga_path") or os.path.join(DATA_DIR, "pga.db")

SITE_URL = sio.read_setting("website") or "https://lutris.net"

INSTALLER_URL = SITE_URL + "/api/installers/%s"

INSTALLER_REVISION_URL = SITE_URL + "/api/installers/game/%s/revisions/%s"
GAME_URL = SITE_URL + "/games/%s/"
RUNTIME_URL = SITE_URL + "/api/runtimes"

STEAM_API_KEY = sio.read_setting("steam_api_key") or "34C9698CEB394AB4401D65927C6B3752"
STEAM_FAMILY_INCLUDE_OWN = sio.read_setting("steam_family_include_own", default="False")

SHOW_MEDIA = os.environ.get("LUTRIS_HIDE_MEDIA") != "1" and sio.read_setting("hide_media") != "True"

DEFAULT_RESOLUTION_WIDTH = sio.read_setting("default_resolution_width", default="1280")
DEFAULT_RESOLUTION_HEIGHT = sio.read_setting("default_resolution_height", default="720")

DEFAULT_USER_AGENT = "Mozilla/5.0 (X11; Linux x86_64; rv:128.0) Gecko/20100101 Firefox/128.0"


read_setting = sio.read_setting
read_bool_setting = sio.read_bool_setting
write_setting = sio.write_setting
SETTINGS_CHANGED = sio.SETTINGS_CHANGED


def get_lutris_directory_settings(directory: str) -> Dict[str, str]:
    """Reads the 'lutris.json' file in 'directory' and returns it as
    a (new) dictionary. The file is missing, unreadable, unparseable, or not a dict,
    this returns an empty dict instead."""
    if directory:
        path = os.path.join(directory, "lutris.json")
        try:
            if os.path.isfile(path):
                with open(path, "r", encoding="utf-8") as f:
                    json_data = json.load(f)
                    if not isinstance(json_data, dict):
                        logger.error("'%s' does not contain a dict, and will be ignored.", path)
                    return cast(Dict[str, str], json_data)
        except Exception as ex:
            logger.exception("Failed to read '%s': %s", path, ex)
    return {}


def set_lutris_directory_settings(directory: str, settings: Dict[str, str], merge: bool = True) -> bool:
    """Updates the 'lutris.json' file in the 'directory' given. If it does not exist, this method creates it. By
    default, if it does exist this merges the values of settings into it, but in a shallow way - only the top level
    entries are merged, not the content any of any sub-dictionaries. If 'merge' is False, this replaces the existing
    'lutris.json', so that its settings are lost.

    This function provides no way to remove a key; you can store nulls instead if appropriate.\

    If this 'lutris.json' file can't be updated (say, if 'settings' can't be represented) then this
    logs the errors, but then returns False."""
    path = os.path.join(directory, "lutris.json")
    if merge and os.path.isfile(path):
        old_settings = get_lutris_directory_settings(directory)
        old_settings.update(settings)
        settings = old_settings

    # In case settings contains something that can't be made into JSON
    # we'll save to a temporary file and rename.
    temp_path = os.path.join(directory, "lutris.json.tmp")
    try:
        with open(temp_path, "w", encoding="utf-8") as f:
            f.write(json.dumps(settings, indent=2))
        os.rename(temp_path, path)
        return True
    except Exception as ex:
        logger.exception("Could not update '%s': %s", path, ex)
        return False
    finally:
        if os.path.isfile(temp_path):
            os.unlink(temp_path)



from gettext import gettext as _

from gi.repository import Gtk

from lutris import settings
from lutris.api import disconnect, read_user_info
from lutris.gui.config.base_config_box import BaseConfigBox
from lutris.gui.config.updates_box import UpdateButtonBox
from lutris.gui.dialogs import ClientLoginDialog, QuestionDialog
from lutris.gui.widgets import EMPTY_NOTIFICATION_REGISTRATION
from lutris.services.lutris import sync_media
from lutris.util.jobs import AsyncCall
from lutris.util.library_sync import (
    LOCAL_LIBRARY_SYNCED,
    LOCAL_LIBRARY_SYNCING,
    LibrarySyncer,
    is_local_library_syncing,
)
from lutris.util.steam.config import STEAM_ACCOUNT_SETTING, get_steam_users
from lutris.util.strings import time_ago


class AccountsBox(BaseConfigBox):
    def __init__(self):
        super().__init__()
        self.add(self.get_section_label(_("Lutris")))
        frame = Gtk.Frame(visible=True, shadow_type=Gtk.ShadowType.ETCHED_IN)
        frame.get_style_context().add_class("info-frame")
        self.bullshit_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, visible=True)
        self.pack_start(frame, False, False, 0)

        self.lutris_options = self.get_lutris_options()
        self.bullshit_box.add(self.lutris_options)
        frame.add(self.bullshit_box)

        self.library_syncing_registration = EMPTY_NOTIFICATION_REGISTRATION
        self.library_synced_registration = EMPTY_NOTIFICATION_REGISTRATION

        self.sync_box = UpdateButtonBox(self.get_sync_box_label(), _("Sync Again"), clicked=self.on_sync_again_clicked)

        self.connect("realize", self.on_realize)
        self.connect("unrealize", self.on_unrealize)

        self.sync_frame = self._get_framed_options_list_box([self.sync_box])
        self.sync_frame.set_visible(settings.read_bool_setting("library_sync_enabled", True))

        self.pack_start(self.sync_frame, False, False, 0)

        self.add(self.get_section_label(_("Steam accounts")))
        self.add(
            self.get_description_label(
                _("Select which Steam account is used for Lutris integration and creating Steam shortcuts.")
            )
        )
        self.frame = Gtk.Frame(visible=True, shadow_type=Gtk.ShadowType.ETCHED_IN)
        self.frame.get_style_context().add_class("info-frame")
        self.pack_start(self.frame, False, False, 0)

        self.accounts_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6, visible=True)
        self.frame.add(self.accounts_box)

    def on_realize(self, _widget):
        self.library_syncing_registration = LOCAL_LIBRARY_SYNCING.register(self.on_local_library_syncing)
        self.library_synced_registration = LOCAL_LIBRARY_SYNCED.register(self.on_local_library_synced)
        if is_local_library_syncing():
            self.on_local_library_syncing()

    def on_unrealize(self, _widget) -> None:
        # The destroy signal never fires for this sub-widget, so we use
        # realize/unrealize for this instead.
        self.library_syncing_registration.unregister()
        self.library_synced_registration.unregister()

    def space_widget(self, widget, top=16, bottom=16):
        widget.set_margin_top(top)
        widget.set_margin_start(16)
        widget.set_margin_end(16)
        widget.set_margin_bottom(bottom)
        return widget

    def get_user_box(self):
        user_info = read_user_info()

        user_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6, visible=True)

        label = Gtk.Label(visible=True)
        label.set_alignment(0, 0.5)
        if user_info:
            label.set_markup(_("Connected as <b>%s</b>") % user_info["username"])
        else:
            label.set_markup(_("Not connected"))
        self.space_widget(label)
        user_box.pack_start(label, True, True, 0)

        if user_info:
            button_text = _("Logout")
            button_handler = self.on_logout_clicked
        else:
            button_text = _("Login")
            button_handler = self.on_login_clicked
        button = Gtk.Button(button_text, visible=True)
        button.connect("clicked", button_handler)
        self.space_widget(button)
        user_box.pack_start(button, False, False, 0)
        return user_box

    def get_lutris_options(self):
        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6, visible=True)
        box.add(self.get_user_box())

        sync_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6, visible=True)
        sync_label = Gtk.Label(_("Keep your game library synced with Lutris.net"), visible=True)
        sync_switch = Gtk.Switch(visible=True)
        sync_switch.set_active(settings.read_bool_setting("library_sync_enabled", default=True))
        sync_switch.connect("state-set", self.on_sync_state_set)
        sync_box.pack_start(sync_label, False, False, 0)
        sync_box.pack_end(sync_switch, False, False, 0)
        self.space_widget(sync_box, bottom=0)
        box.add(sync_box)

        label = Gtk.Label(visible=True)
        label.set_alignment(0, 0.5)
        label.set_markup(
            _(
                "<i>This will send play time, last played, runner, platform \n"
                "and store information to the lutris website so you can \n"
                "sync this data on multiple devices</i>"
            )
        )
        self.space_widget(label, top=0)
        box.add(label)

        return box

    def populate_steam_accounts(self):
        main_radio_button = None
        active_steam_account = settings.read_setting(STEAM_ACCOUNT_SETTING)

        steam_users = get_steam_users()
        for account in steam_users:
            steamid64 = account["steamid64"]
            name = account.get("PersonaName") or f"#{steamid64}"
            radio_button = Gtk.RadioButton.new_with_label_from_widget(main_radio_button, name)
            self.space_widget(radio_button)
            radio_button.show()
            radio_button.set_active(active_steam_account == steamid64)
            radio_button.connect("toggled", self.on_steam_account_toggled, steamid64)
            self.accounts_box.pack_start(radio_button, True, True, 0)
            if not main_radio_button:
                main_radio_button = radio_button
        if not steam_users:
            self.accounts_box.pack_start(
                self.space_widget(Gtk.Label(_("No Steam account found"), visible=True)),
                True,
                True,
                0,
            )

    def rebuild_lutris_options(self):
        self.bullshit_box.remove(self.lutris_options)
        self.lutris_options.destroy()
        self.lutris_options = self.get_lutris_options()
        self.bullshit_box.add(self.lutris_options)

    def on_logout_clicked(self, _widget):
        disconnect()
        self.rebuild_lutris_options()

    def on_login_clicked(self, _widget):
        login_dialog = ClientLoginDialog(parent=self.get_toplevel())
        login_dialog.connect("connected", self.on_connect_response)

    def on_connect_response(self, _dialog, bliblu):
        self.rebuild_lutris_options()

    def on_sync_again_clicked(self, _button):
        AsyncCall(LibrarySyncer().sync_local_library, None, force=True)

    def on_local_library_syncing(self):
        self.sync_box.show_running_markup(_("<i>Syncing library...</i>"))

    def on_local_library_synced(self):
        self.sync_box.show_completion_markup(self.get_sync_box_label(), "")
        AsyncCall(sync_media, None)

    def get_sync_box_label(self):
        synced_at = settings.read_setting("last_library_sync_at")
        if synced_at:
            return _("<i>Last synced %s.</i>") % time_ago(int(synced_at))

        return ""

    def on_steam_account_toggled(self, radio_button, steamid64):
        """Handler for switching the active Steam account."""
        settings.write_setting(STEAM_ACCOUNT_SETTING, steamid64)

    def on_sync_state_set(self, switch, state):
        if not settings.read_setting("last_library_sync_at"):
            sync_warn_dialog = QuestionDialog(
                {
                    "title": _("Synchronize library?"),
                    "question": _("Enable library sync and run a full sync with lutris.net?"),
                }
            )
            if sync_warn_dialog.result == Gtk.ResponseType.YES:
                AsyncCall(LibrarySyncer().sync_local_library, None)
            else:
                return

        self.on_setting_change(switch, state, "library_sync_enabled")
        self.sync_frame.set_visible(state)



"""Ubisoft Connect service"""

import glob
import json
import os
import shutil
from gettext import gettext as _
from typing import Any, Dict, Optional
from urllib.parse import unquote

from gi.repository import Gio

from lutris import settings
from lutris.config import LutrisConfig, write_game_config
from lutris.database.games import add_game, get_game_by_field, update_existing
from lutris.database.services import ServiceGameCollection
from lutris.game import Game
from lutris.services.base import SERVICE_LOGIN, SERVICE_LOGOUT, OnlineService
from lutris.services.lutris import sync_media
from lutris.services.service_game import ServiceGame
from lutris.services.service_media import ServiceMedia
from lutris.util.jobs import AsyncCall
from lutris.util.log import logger
from lutris.util.strings import slugify
from lutris.util.ubisoft import consts
from lutris.util.ubisoft.client import UbisoftConnectClient
from lutris.util.ubisoft.helpers import get_ubisoft_registry
from lutris.util.ubisoft.parser import UbisoftParser
from lutris.util.wine.prefix import WinePrefixManager


class UbisoftCover(ServiceMedia):
    """Ubisoft connect cover art"""

    service = "ubisoft"
    size = (160, 186)
    dest_path = os.path.join(settings.CACHE_DIR, "ubisoft/covers")
    file_patterns = ["%s.jpg"]
    api_field = "thumbImage"
    url_pattern = "https://static3.cdn.ubi.com/orbit/uplay_launcher_3_0/assets/%s"

    def get_media_url(self, details: Dict[str, Any]) -> Optional[str]:
        # First try coverUrl from the API (available for games fetched via GraphQL)
        if details.get("coverUrl"):
            return details["coverUrl"]
        # Fall back to thumbImage from local config files (for locally parsed games)
        if details.get(self.api_field):
            return self.url_pattern % details[self.api_field]
        # No image available - this is expected for some games
        return None

    def download(self, slug, url):
        if url.startswith("http"):
            return super().download(slug, url)
        if not url.endswith(".jpg"):
            return
        ubi_game = get_game_by_field("ubisoft-connect", "slug")
        if not ubi_game:
            return
        base_dir = ubi_game["directory"]
        asset_file = os.path.join(
            base_dir, "drive_c/Program Files (x86)/Ubisoft/Ubisoft Game Launcher/cache/assets", url
        )
        cache_path = os.path.join(self.dest_path, self.get_filename(slug))
        if os.path.exists(asset_file):
            shutil.copy(asset_file, cache_path)
        else:
            logger.warning("No thumbnail in %s", asset_file)


class UbisoftGame(ServiceGame):
    """Service game for games from Ubisoft connect"""

    service = "ubisoft"

    @classmethod
    def new_from_api(cls, payload):
        """Convert an Ubisoft game to a service game"""
        service_game = cls()
        service_game.appid = payload["spaceId"] or payload["installId"]
        service_game.slug = slugify(payload["name"])
        service_game.name = payload["name"]
        service_game.details = json.dumps(payload)
        return service_game


class UbisoftConnectService(OnlineService):
    """Service class for Ubisoft Connect"""

    id = "ubisoft"
    name = _("Ubisoft Connect")
    icon = "ubisoft"
    runner = "wine"
    client_installer = "ubisoft-connect"
    browser_size = (460, 690)
    login_window_width = 460
    login_window_height = 690
    cookies_path = os.path.join(settings.CACHE_DIR, "ubisoft/.auth")
    token_path = os.path.join(settings.CACHE_DIR, "ubisoft/.token")
    cache_path = os.path.join(settings.CACHE_DIR, "ubisoft/library/")
    login_url = consts.LOGIN_URL
    redirect_uris = ["https://connect.ubisoft.com/change_domain/"]
    scripts = {
        "https://connect.ubisoft.com/ready": ('window.location.replace("https://connect.ubisoft.com/change_domain/");'),
        "https://connect.ubisoft.com/change_domain/": (
            'window.location.replace(localStorage.getItem("PRODloginData") +","+ '
            'localStorage.getItem("PRODrememberMe") +"," + localStorage.getItem("PRODlastProfile"));'
        ),
    }
    medias = {
        "cover": UbisoftCover,
    }
    default_format = "cover"

    def __init__(self):
        super().__init__()
        self.client = UbisoftConnectClient(self)

    def auth_lost(self):
        SERVICE_LOGOUT.fire(self)

    def login_callback(self, credentials):
        """Called after the user has logged in successfully"""
        logger.info("Login to Ubisoft Connect sucessful")
        url = credentials[len("https://connect.ubisoft.com/change_domain/") :]
        unquoted_url = unquote(url)
        storage_jsons = json.loads("[" + unquoted_url + "]")
        user_data = self.client.authorise_with_local_storage(storage_jsons)
        logger.debug("Ubisoft user data: %s", user_data)
        self.client.set_auth_lost_callback(self.auth_lost)
        SERVICE_LOGIN.fire(self)
        return (user_data["userId"], user_data["username"])

    def is_connected(self):
        res = self.is_authenticated()
        logger.debug("Ubisoft Connect is connected: %s", res)
        return res

    def get_configurations(self):
        ubi_game = get_game_by_field("ubisoft-connect", "slug")
        if not ubi_game:
            return None

        base_dir = ubi_game["directory"]

        """Define potential relative paths for configuration files across different launcher versions"""
        possible_paths = [
            "drive_c/Program Files (x86)/Ubisoft/Ubisoft Game Launcher/cache/configuration/configurations",
            "drive_c/users/*/AppData/Local/Ubisoft Game Launcher/cache/configuration/configurations",
        ]

        all_valid_files = []

        """Iterate through patterns to find all matching files on the system"""
        for path_pattern in possible_paths:
            full_pattern = os.path.join(base_dir, path_pattern)
            matching_paths = glob.glob(full_pattern)

            """Filter results to ensure we only collect actual files (ignoring directories)"""
            for p in matching_paths:
                if os.path.isfile(p):
                    all_valid_files.append(p)

        """Log all discovered candidates for debugging purposes"""
        for f in all_valid_files:
            logger.debug("Found config file candidate: %s", f)

        """If any valid files were found, select and read the most recently modified one"""
        if all_valid_files:
            latest_file = max(all_valid_files, key=os.path.getmtime)

            logger.debug("Loading configuration from: %s", latest_file)
            with open(latest_file, "rb") as f:
                return f.read()

        """Fallback if no configuration files were detected"""
        logger.info("Ubisoft configuration file not found in %s", base_dir)
        return None

    def load(self):
        try:
            self.client.authorise_with_stored_credentials(self.load_credentials())
        except RuntimeError as ex:
            logger.error("Failed to authorize with API: %s. Re-login required." % ex)
            AsyncCall(self.logout, lambda _result, _error: self.login())
            return
        response = self.client.get_club_titles()
        games = response["data"]["viewer"]["ownedGames"].get("nodes", [])
        ubi_games = []
        for game in games:
            if "ownedPlatformGroups" in game:
                is_pc = False
                for platform_group in game["ownedPlatformGroups"]:
                    for platform in platform_group:
                        if platform["type"] == "PC":
                            is_pc = True
                if not is_pc:
                    continue
            ubi_game = UbisoftGame.new_from_api(game)
            ubi_game.save()
            ubi_games.append(ubi_game)
        configuration_data = self.get_configurations()
        config_parser = UbisoftParser()
        for game in config_parser.parse_games(configuration_data):
            ubi_game = UbisoftGame.new_from_api(game)
            ubi_game.save()
            ubi_games.append(ubi_game)
        return ubi_games

    @property
    def credential_files(self):
        """Return a list of all files used for authentication"""
        return [self.token_path]

    def store_credentials(self, credentials):
        if not os.path.exists(os.path.dirname(self.token_path)):
            logger.debug("Creating Ubisoft credentials path: %s", self.token_path)
            os.mkdir(os.path.dirname(self.token_path))

        logger.debug("Writing Ubisoft credentials to %s", self.token_path)
        with open(self.token_path, "w", encoding="utf-8") as auth_file:
            auth_file.write(json.dumps(credentials, indent=2))

    def load_credentials(self):
        logger.debug("Loading credentials from %s", self.token_path)
        with open(self.token_path, encoding="utf-8") as auth_file:
            credentials = json.load(auth_file)
        return credentials

    def install_from_ubisoft(self, ubisoft_connect, game):
        app_name = game["name"]

        lutris_game_id = slugify(game["name"]) + "-" + self.id
        existing_game = get_game_by_field(lutris_game_id, "installer_slug")
        if existing_game and existing_game["installed"] == 1:
            logger.debug("Ubisoft Connect game %s installed in Lutris", app_name)
            return
        logger.debug("Installing Ubisoft Connect game %s", app_name)
        game_config = LutrisConfig(game_config_id=ubisoft_connect["configpath"]).game_level
        details = json.loads(game["details"])
        launch_id = details.get("launchId") or details.get("installId") or details.get("spaceId")
        game_config["game"]["args"] = f"uplay://launch/{launch_id}"
        configpath = write_game_config(lutris_game_id, game_config)
        slug = self.get_installed_slug(game)
        if existing_game:
            update_existing(
                id=existing_game["id"],
                name=game["name"],
                runner=self.runner,
                slug=slug,
                directory=ubisoft_connect["directory"],
                installed=1,
                installer_slug=lutris_game_id,
                configpath=configpath,
                service=self.id,
                service_id=game["appid"],
            )
            return existing_game["id"]
        add_game(
            name=game["name"],
            runner=self.runner,
            slug=slug,
            directory=ubisoft_connect["directory"],
            installed=1,
            installer_slug=lutris_game_id,
            configpath=configpath,
            service=self.id,
            service_id=game["appid"],
        )
        return slug

    def add_installed_games(self):
        ubisoft_connect = get_game_by_field(self.client_installer, "slug")
        if not ubisoft_connect:
            logger.warning("Ubisoft Connect not installed")
            return
        prefix_path = ubisoft_connect["directory"].split("drive_c")[0]
        prefix = WinePrefixManager(prefix_path)
        installed_slugs = []
        for game in ServiceGameCollection.get_for_service(self.id):
            details = json.loads(game["details"])
            install_path = get_ubisoft_registry(prefix, details.get("registryPath"))
            exe = get_ubisoft_registry(prefix, details.get("exe"))
            if install_path and exe:
                slug = self.install_from_ubisoft(ubisoft_connect, game)
                if slug:
                    installed_slugs.append(slug)
        logger.debug("Syncing media for %s games", len(installed_slugs))
        sync_media(installed_slugs)

    def generate_installer(self, db_game, ubi_db_game):
        ubisoft_connect = Game(ubi_db_game["id"])
        uc_exe = ubisoft_connect.config.game_config["exe"]
        if not os.path.isabs(uc_exe):
            uc_exe = os.path.join(ubisoft_connect.config.game_config["prefix"], uc_exe)
        details = json.loads(db_game["details"])
        launch_id = details.get("launchId") or details.get("installId") or details.get("spaceId")
        install_id = details.get("installId") or details.get("launchId") or details.get("spaceId")
        return {
            "name": db_game["name"],
            "version": self.name,
            "slug": slugify(db_game["name"]) + "-" + self.id,
            "game_slug": self.get_installed_slug(db_game),
            "runner": self.get_installed_runner_name(db_game),
            "appid": db_game["appid"],
            "script": {
                "requires": self.client_installer,
                "game": {
                    "args": f"uplay://launch/{launch_id}",
                },
                "installer": [
                    {
                        "task": {
                            "name": "wineexec",
                            "executable": uc_exe,
                            "args": f"uplay://install/{install_id}",
                            "prefix": ubisoft_connect.config.game_config["prefix"],
                            "description": (
                                "Ubisoft will now open and install %s. "
                                "Close Ubisoft Connect to complete the install process."
                            )
                            % db_game["name"],
                        }
                    }
                ],
            },
        }

    def get_installed_runner_name(self, db_game):
        return self.runner

    def install(self, db_game):
        """Install a game or Ubisoft Connect if not already installed"""
        ubisoft_connect = get_game_by_field(self.client_installer, "slug")
        application = Gio.Application.get_default()
        if not ubisoft_connect or not ubisoft_connect["installed"]:
            logger.warning("Ubisoft Connect (%s) not installed", self.client_installer)
            application.show_lutris_installer_window(game_slug=self.client_installer)
        else:
            application.show_installer_window(
                [self.generate_installer(db_game, ubisoft_connect)], service=self, appid=db_game["appid"]
            )



"""Service game module"""

from lutris import settings
from lutris.database import sql
from lutris.database.services import ServiceGameCollection
from lutris.services.service_media import ServiceMedia


class ServiceGame:
    """Representation of a game from a 3rd party service"""

    service = NotImplemented
    installer_slug = NotImplemented
    medias = (ServiceMedia,)

    def __init__(self):
        self.appid = None  # External ID of the game on the 3rd party service
        self.game_id = None  # Internal Lutris ID
        self.runner = None  # Name of the runner
        self.name = None  # Name
        self.slug = None  # Game slug
        self.lutris_slug = None  # Slug used by the lutris website
        self.logo = None  # Game logo
        self.icon = None  # Game icon
        self.details = None  # Additional details for the game

    def save(self):
        """Save this game to database"""
        game_data = {
            "service": self.service,
            "appid": self.appid,
            "name": self.name,
            "slug": self.slug,
            "lutris_slug": self.lutris_slug,
            "icon": self.icon,
            "logo": self.logo,
            "details": str(self.details),
        }
        existing_game = ServiceGameCollection.get_game(self.service, self.appid)
        if existing_game:
            sql.db_update(settings.DB_PATH, "service_games", game_data, {"id": existing_game["id"]})
        else:
            sql.db_insert(settings.DB_PATH, "service_games", game_data)



"""Transform runner parameters to data usable for runtime execution"""

import os
import shlex
import stat

from lutris.util import cache_single, system
from lutris.util.graphics.gpu import GPUS
from lutris.util.linux import LINUX_SYSTEM
from lutris.util.log import logger


def get_mangohud_conf(system_config):
    """Return correct launch arguments and environment variables for Mangohud."""
    # The environment variable should be set to 0 on gamescope, otherwise the game will crash
    if system_config.get("gamescope"):
        return None, None
    if system_config.get("mangohud") and system.can_find_executable("mangohud"):
        return ["mangohud"], {"MANGOHUD": "1", "MANGOHUD_DLSYM": "1"}
    return None, None


def get_launch_parameters(runner, gameplay_info):
    system_config = runner.system_config
    launch_arguments = gameplay_info["command"]
    env = {}

    # MangoHud
    if runner.name == "steam":
        logger.info(
            "Do not enable Mangodhud for Steam games in Lutris."
            "Edit the launch options in Steam and set them to mangohud %%command%%"
        )
    else:
        mango_args, mango_env = get_mangohud_conf(system_config)
        if mango_args:
            launch_arguments = mango_args + launch_arguments
            env.update(mango_env)

    prefix_command = system_config.get("prefix_command") or ""
    if prefix_command:
        launch_arguments = shlex.split(os.path.expandvars(prefix_command)) + launch_arguments

    single_cpu = system_config.get("single_cpu") or False
    if single_cpu:
        limit_cpu_count = system_config.get("limit_cpu_count")
        if limit_cpu_count and limit_cpu_count.isnumeric():
            limit_cpu_count = int(limit_cpu_count)
        else:
            limit_cpu_count = 1

        limit_cpu_count = max(1, limit_cpu_count)
        logger.info("The game will run on %d CPU core(s)", limit_cpu_count)
        launch_arguments.insert(0, "0-%d" % (limit_cpu_count - 1))
        launch_arguments.insert(0, "-c")
        launch_arguments.insert(0, "taskset")

    env.update(runner.get_env())

    env.update(gameplay_info.get("env") or {})

    # Set environment variables dependent on gameplay info

    # LD_PRELOAD
    ld_preload = gameplay_info.get("ld_preload")
    if ld_preload:
        env["LD_PRELOAD"] = ld_preload

    # LD_LIBRARY_PATH
    game_ld_library_path = gameplay_info.get("ld_library_path")
    if game_ld_library_path:
        ld_library_path = env.get("LD_LIBRARY_PATH")
        env["LD_LIBRARY_PATH"] = os.pathsep.join(filter(None, [game_ld_library_path, ld_library_path]))

    # Feral gamemode
    gamemode = system_config.get("gamemode") and LINUX_SYSTEM.gamemode_available()
    if gamemode:
        launch_arguments.insert(0, "gamemoderun")

    # Gamescope
    has_gamescope = system_config.get("gamescope") and system.can_find_executable("gamescope")
    if has_gamescope:
        launch_arguments = get_gamescope_args(launch_arguments, system_config)
        if system_config.get("gamescope_hdr"):
            env["DXVK_HDR"] = "1"

    return launch_arguments, env


def get_gamescope_args(launch_arguments, system_config):
    """Insert gamescope at the start of the launch arguments"""
    launch_arguments.insert(0, "--")
    if system_config.get("gamescope_force_grab_cursor"):
        launch_arguments.insert(0, "--force-grab-cursor")
    if system_config.get("gamescope_fsr_sharpness"):
        launch_arguments.insert(0, system_config["gamescope_fsr_sharpness"])
        launch_arguments.insert(0, "--fsr-sharpness")
        launch_arguments[0:0] = _get_gamescope_fsr_option()
    if system_config.get("gamescope_flags"):
        gamescope_flags = shlex.split(system_config["gamescope_flags"])
        launch_arguments = gamescope_flags + launch_arguments
    if system_config.get("gamescope_window_mode"):
        gamescope_window_mode = system_config["gamescope_window_mode"]
        launch_arguments.insert(0, gamescope_window_mode)
    if system_config.get("gamescope_fps_limiter"):
        gamescope_fps_limiter = system_config["gamescope_fps_limiter"]
        launch_arguments.insert(0, gamescope_fps_limiter)
        launch_arguments.insert(0, "-r")
    if system_config.get("gamescope_output_res"):
        output_width, output_height = system_config["gamescope_output_res"].lower().split("x")
        launch_arguments.insert(0, output_height)
        launch_arguments.insert(0, "-H")
        launch_arguments.insert(0, output_width)
        launch_arguments.insert(0, "-W")
    if system_config.get("gamescope_game_res"):
        game_width, game_height = system_config["gamescope_game_res"].lower().split("x")
        launch_arguments.insert(0, game_height)
        launch_arguments.insert(0, "-h")
        launch_arguments.insert(0, game_width)
        launch_arguments.insert(0, "-w")
    if len(GPUS) > 1 and system_config.get("gpu") in GPUS:
        gpu = GPUS[system_config["gpu"]]
        launch_arguments.insert(0, gpu.pci_id)
        launch_arguments.insert(0, "--prefer-vk-device")
    if system_config.get("gamescope_hdr"):
        launch_arguments.insert(0, "--hdr-enabled")
    if system_config.get("mangohud"):
        launch_arguments.insert(0, "--mangoapp")
    launch_arguments.insert(0, "gamescope")
    return launch_arguments


@cache_single
def _get_gamescope_fsr_option():
    """Returns a list containing the arguments to insert to trigger FSR in gamescope;
    this changes in later versions, so we have to check the help output. There seems to be
    no way to query the version number more directly."""
    if system.can_find_executable("gamescope"):
        # '-F fsr' is the trigger in gamescope 3.12.
        stdout, stderr = system.execute_with_error(["gamescope", "--help"])
        help_text = stdout + stderr
        if "-F, --filter" in help_text:
            return ["-F", "fsr"]

    # This is the old trigger, pre 3.12.
    return ["-U"]


def export_bash_script(runner, gameplay_info, script_path):
    """Convert runner configuration into a bash script"""
    runner.prelaunch()
    command, env = get_launch_parameters(runner, gameplay_info)
    # Override TERM otherwise the script might not run
    env["TERM"] = "xterm"
    script_content = "#!/bin/bash\n\n\n"

    script_content += "# Environment variables\n"

    for name, value in env.items():
        script_content += 'export %s="%s"\n' % (name, value)

    if "working_dir" in gameplay_info:
        script_content += "\n# Working Directory\n"
        script_content += "cd %s\n" % shlex.quote(gameplay_info["working_dir"])

    script_content += "\n# Command\n"
    script_content += " ".join([shlex.quote(c) for c in command])

    with open(script_path, "w", encoding="utf-8") as script_file:
        script_file.write(script_content)

    os.chmod(script_path, os.stat(script_path).st_mode | stat.S_IEXEC)



from lutris import settings
from lutris.database import sql
from lutris.util.log import logger

DATABASE = {
    "games": [
        {"name": "id", "type": "INTEGER", "indexed": True},
        {"name": "name", "type": "TEXT"},
        {
            "name": "sortname",
            "type": "TEXT",
        },
        {"name": "slug", "type": "TEXT"},
        {"name": "installer_slug", "type": "TEXT"},
        {"name": "parent_slug", "type": "TEXT"},
        {"name": "platform", "type": "TEXT"},
        {"name": "runner", "type": "TEXT"},
        {"name": "executable", "type": "TEXT"},
        {"name": "directory", "type": "TEXT"},
        {"name": "updated", "type": "DATETIME"},
        {"name": "lastplayed", "type": "INTEGER"},
        {"name": "installed", "type": "INTEGER"},
        {"name": "installed_at", "type": "INTEGER"},
        {"name": "year", "type": "INTEGER"},
        {"name": "configpath", "type": "TEXT"},
        {"name": "has_custom_banner", "type": "INTEGER"},
        {"name": "has_custom_icon", "type": "INTEGER"},
        {"name": "has_custom_coverart_big", "type": "INTEGER"},
        {"name": "playtime", "type": "REAL"},
        {"name": "service", "type": "TEXT"},
        {"name": "service_id", "type": "TEXT"},
        {
            "name": "discord_id",
            "type": "TEXT",
        },
    ],
    "service_games": [
        {"name": "id", "type": "INTEGER", "indexed": True},
        {"name": "service", "type": "TEXT"},
        {"name": "appid", "type": "TEXT"},
        {"name": "name", "type": "TEXT"},
        {"name": "slug", "type": "TEXT"},
        {"name": "icon", "type": "TEXT"},
        {"name": "logo", "type": "TEXT"},
        {"name": "url", "type": "TEXT"},
        {"name": "details", "type": "TEXT"},
        {"name": "lutris_slug", "type": "TEXT"},
    ],
    "sources": [
        {"name": "id", "type": "INTEGER", "indexed": True},
        {"name": "uri", "type": "TEXT UNIQUE"},
    ],
    "categories": [
        {"name": "id", "type": "INTEGER", "indexed": True},
        {"name": "name", "type": "TEXT", "unique": True},
    ],
    "games_categories": [
        {"name": "game_id", "type": "INTEGER", "indexed": False},
        {"name": "category_id", "type": "INTEGER", "indexed": False},
    ],
    "saved_searches": [
        {"name": "id", "type": "INTEGER", "indexed": True},
        {"name": "name", "type": "TEXT", "unique": True},
        {"name": "search", "type": "TEXT", "unique": False},
    ],
}


def get_schema(tablename):
    """
    Fields:
        - position
        - name
        - type
        - not null
        - default
        - indexed
    """
    tables = []
    query = "pragma table_info('%s')" % tablename
    with sql.db_cursor(settings.DB_PATH) as cursor:
        for row in cursor.execute(query).fetchall():
            field = {
                "name": row[1],
                "type": row[2],
                "not_null": row[3],
                "default": row[4],
                "indexed": row[5],
            }
            tables.append(field)
    return tables


def field_to_string(name="", type="", indexed=False, unique=False):  # pylint: disable=redefined-builtin
    """Converts a python based table definition to it's SQL statement"""
    field_query = "%s %s" % (name, type)
    if indexed:
        field_query += " PRIMARY KEY"
    if unique:
        field_query += " UNIQUE"
    return field_query


def create_table(name, schema):
    """Creates a new table in the database"""
    fields = ", ".join([field_to_string(**f) for f in schema])
    query = "CREATE TABLE IF NOT EXISTS %s (%s)" % (name, fields)
    logger.debug("[Query] %s", query)
    with sql.db_cursor(settings.DB_PATH) as cursor:
        cursor.execute(query)


def migrate(table, schema):
    """Compare a database table with the reference model and make necessary changes

    This is very basic and only the needed features have been implemented (adding columns)

    Args:
        table (str): Name of the table to migrate
        schema (dict): Reference schema for the table

    Returns:
        list: The list of column names that have been added
    """

    existing_schema = get_schema(table)
    migrated_fields = []
    if existing_schema:
        columns = [col["name"] for col in existing_schema]
        for field in schema:
            if field["name"] not in columns:
                logger.info("Migrating %s field %s", table, field["name"])
                migrated_fields.append(field["name"])
                sql.add_field(settings.DB_PATH, table, field)
    else:
        create_table(table, schema)
    return migrated_fields


def syncdb():
    """Update the database to the current version, making necessary changes
    for backwards compatibility."""
    for table_name, table_data in DATABASE.items():
        migrate(table_name, table_data)



"""Grid view for the main window"""

# pylint: disable=no-member
from gi.repository import Gtk

from lutris import settings
from lutris.gui.views import COL_ID, COL_INSTALLED, COL_MEDIA_PATHS, COL_NAME, COL_PLATFORM
from lutris.gui.views.base import GameView
from lutris.gui.widgets.cellrenderers import GridViewCellRendererImage, GridViewCellRendererText
from lutris.util.log import logger


class GameGridView(Gtk.IconView, GameView):  # type:ignore[misc]
    __gsignals__ = GameView.__gsignals__

    min_width = 70  # Minimum width for a cell

    def __init__(self, store, hide_text=False):
        Gtk.IconView.__init__(self)
        GameView.__init__(self)

        Gtk.IconView.set_selection_mode(self, Gtk.SelectionMode.MULTIPLE)

        self.set_column_spacing(6)
        self._show_badges = True

        if settings.SHOW_MEDIA:
            self.image_renderer = GridViewCellRendererImage()
            self.pack_start(self.image_renderer, False)
            self._initialize_image_renderer_attributes()
        else:
            self.image_renderer = None
        self.set_item_padding(1)
        if hide_text:
            self.text_renderer = None
        else:
            self.text_renderer = GridViewCellRendererText()
            self.pack_end(self.text_renderer, False)
            self.add_attribute(self.text_renderer, "markup", COL_NAME)

        self.set_game_store(store)

        self.connect_signals()
        self.connect("item-activated", self.on_item_activated)
        self.connect("selection-changed", self.on_selection_changed)
        self.connect("style-updated", self.on_style_updated)

    def set_game_store(self, game_store):
        super().set_game_store(game_store)
        self.model = game_store.store
        self.set_model(self.model)

        size = game_store.service_media.size
        if self.image_renderer:
            self.image_renderer.set_expected_size(size[0], size[1])
        if self.text_renderer:
            cell_width = max(size[0], self.min_width)
            self.text_renderer.set_width(cell_width)

    @property
    def show_badges(self):
        return self._show_badges

    @show_badges.setter
    def show_badges(self, value):
        if self._show_badges != value:
            self._show_badges = value
            self._initialize_image_renderer_attributes()
            self.queue_draw()

    def _initialize_image_renderer_attributes(self):
        if self.image_renderer:
            self.image_renderer.show_badges = self.show_badges
            self.clear_attributes(self.image_renderer)
            self.add_attribute(self.image_renderer, "game_id", COL_ID)
            self.add_attribute(self.image_renderer, "media_paths", COL_MEDIA_PATHS)
            self.add_attribute(self.image_renderer, "platform", COL_PLATFORM)
            self.add_attribute(self.image_renderer, "is_installed", COL_INSTALLED)

    def get_path_at(self, x, y):
        return self.get_path_at_pos(x, y)

    def set_selected(self, paths, scroll_into_view=False):
        self.unselect_all()

        for idx, path in enumerate(paths):
            self.select_path(path)
            if scroll_into_view and idx == 0:
                self.scroll_to_path(path, False, 0.0, 0.0)

    def get_selected(self):
        """Return list of all selected items"""
        return self.get_selected_items()

    def get_game_id_for_path(self, path):
        iterator = self.get_model().get_iter(path)
        return self.get_model().get_value(iterator, COL_ID)

    def get_path_for_game_id(self, game_id):
        path_found = None

        def check_path(model, path, iterator):
            nonlocal path_found
            row_id = model.get_value(iterator, COL_ID)
            if game_id == row_id:
                path_found = path
                return True  # stop iteration

        self.get_model().foreach(check_path)
        return path_found

    def on_item_activated(self, _view, _path):
        """Handles double clicks"""
        selected_id = self.get_selected_game_id()
        if selected_id:
            logger.debug("Item activated: %s", selected_id)
            self.emit("game-activated", selected_id)

    def on_selection_changed(self, _view):
        """Handles selection changes"""
        selected_items = self.get_selected()
        self.emit("game-selected", selected_items)

    def on_style_updated(self, widget):
        if self.text_renderer:
            self.text_renderer.clear_caches()



from lutris import settings
from lutris.database import sql
from lutris.util.log import logger


class ServiceGameCollection:
    @classmethod
    def get_service_games(cls, searches=None, filters=None, excludes=None, sorts=None):
        return sql.filtered_query(
            settings.DB_PATH, "service_games", searches=searches, filters=filters, excludes=excludes, sorts=sorts
        )

    @classmethod
    def get_for_service(cls, service):
        if not service:
            raise ValueError("No service provided")
        return sql.filtered_query(settings.DB_PATH, "service_games", filters={"service": service})

    @classmethod
    def get_game(cls, service, appid):
        """Return a single game referred by its appid"""
        if not service:
            raise ValueError("No service provided")
        if not appid:
            raise ValueError("No appid provided")
        results = sql.filtered_query(settings.DB_PATH, "service_games", filters={"service": service, "appid": appid})
        if not results:
            return
        if len(results) > 1:
            logger.warning("More than one game found for %s on %s", appid, service)
        return results[0]



import sys
import threading
import traceback
from typing import Callable, Optional

from gi.repository import GLib  # type: ignore

from lutris.util.log import logger


class AsyncCall(threading.Thread):
    def __init__(self, func, callback, *args, callback_target=None, **kwargs):
        """Execute `function` in a new thread then schedule `callback` for
        execution in the main loop. If 'callback_target' is a widget and it is destroyed
        in the meantime, the callback is cancelled.
        """
        self.callback_task = None
        self.stop_request = threading.Event()

        super().__init__(target=self.target, args=args, kwargs=kwargs)
        self.function = func
        if not callback:
            self.callback = lambda r, e: None
        else:
            self.callback = self._protect_callback(callback, callback_target)
        self.daemon = kwargs.pop("daemon", True)
        self.start()

    def _protect_callback(self, callback, callback_target=None):
        """Wraps and hooks up an on-destroyed handler on the callback_target that
        removes the callback; this prevents an AsyncJob from completing on a
        destroyed widget, which can cause a crash.

        If no callback_target is given, this will use the receiver of the callback
        (if it has one)."""
        if not callback_target:
            callback_target = callback.__self__ if hasattr(callback, "__self__") else None

        if callback_target and hasattr(callback_target, "call_when_destroyed"):

            def unhook():
                # If the target is destroyed, block the callback; no need to disconnect
                # from a dead object.
                self.callback = lambda r, e: None

            def fire(r, e):
                # Before starting the callback, unhook the on-destroyed callback
                # so we don't leak it.
                disconnecter()
                callback(r, e)

            disconnecter = callback_target.call_when_destroyed(unhook)
            return fire
        else:
            return callback

    def target(self, *a, **kw):
        result = None
        error = None

        try:
            result = self.function(*a, **kw)
        except Exception as ex:  # pylint: disable=broad-except
            logger.error("Error while completing task %s: %s %s", self.function, type(ex), ex)
            error = ex
            _ex_type, _ex_value, trace = sys.exc_info()
            traceback.print_tb(trace)

        self.callback_task = schedule_at_idle(self.callback, result, error)


class IdleTask:
    """This class provides a safe interface for cancelling idle tasks and timeouts;
    this will simply do nothing after being used once, and once the task completes,
    it will also do nothing.

    These objects are returned by the schedule methods below, which disconnect
    them when appropriate."""

    def __init__(self) -> None:
        """Initializes a task with no connection to a source, but also not completed; this can be
        connected to a source via the connect() method, unless it is completed first."""
        self.source_id: Optional[int] = None
        self._is_completed = False

    def unschedule(self) -> None:
        """Call this to prevent the idle task from running, if it has not already run."""
        if self.source_id is not None:
            GLib.source_remove(self.source_id)
            self.disconnect()

    def is_completed(self) -> bool:
        """True if the idle task has completed; that is, if mark_completed() was called on it."""
        return self._is_completed

    def connect(self, source_id) -> None:
        """Connects this task to a source to be unscheduled; but if the task is already
        completed, this does nothing."""
        if not self._is_completed:
            self.source_id = source_id

    def disconnect(self) -> None:
        """Break the link to the idle task, so it can't be unscheduled."""
        self.source_id = None

    def mark_completed(self) -> None:
        """Marks the task as completed, and also disconnect it."""
        self._is_completed = True
        self.disconnect()


# A task that is always completed and disconnected and does nothing.
COMPLETED_IDLE_TASK = IdleTask()
COMPLETED_IDLE_TASK.mark_completed()


def schedule_at_idle(func: Callable[..., None], *args, delay_seconds: float = 0.0) -> IdleTask:
    """Schedules a function to run at idle time, once. You can specify a delay in seconds
    before it runs.
    Returns an object to prevent it running."""

    task = IdleTask()

    def wrapper(*a, **kw) -> bool:
        try:
            func(*a, **kw)
            return False
        finally:
            task.disconnect()

    handler_object = func.__self__ if hasattr(func, "__self__") else None
    if handler_object:
        wrapper.__self__ = handler_object  # type: ignore[attr-defined]

    if delay_seconds >= 0.0:
        milliseconds = int(delay_seconds * 1000)
        source_id = GLib.timeout_add(milliseconds, wrapper, *args)
    else:
        source_id = GLib.idle_add(wrapper, *args)

    task.connect(source_id)
    return task


def schedule_repeating_at_idle(
    func: Callable[..., bool],
    *args,
    interval_seconds: float = 0.0,
) -> IdleTask:
    """Schedules a function to run at idle time, over and over until it returns False.
    It can be repeated at an interval in seconds, which will also delay it's first invocation.
    Returns an object to stop it running."""

    task = IdleTask()

    def wrapper(*a, **kw) -> bool:
        repeat = False
        try:
            repeat = func(*a, **kw)
            return repeat
        finally:
            if not repeat:
                task.disconnect()

    handler_object = func.__self__ if hasattr(func, "__self__") else None
    if handler_object:
        wrapper.__self__ = handler_object  # type: ignore[attr-defined]

    if interval_seconds >= 0.0:
        milliseconds = int(interval_seconds * 1000)
        source_id = GLib.timeout_add(milliseconds, wrapper, *args)
    else:
        source_id = GLib.idle_add(wrapper, *args)

    task.connect(source_id)
    return task



import json
import os

from lutris.util import system
from lutris.util.log import logger


def get_gog_game_path(target_path):
    """Return the absolute path where a GOG game is installed"""
    gog_game_path = os.path.join(target_path, "drive_c/GOG Games/")
    if not os.path.exists(gog_game_path):
        logger.warning("No 'GOG Games' folder in %s", target_path)
        return None
    games = os.listdir(gog_game_path)
    if len(games) > 1:
        logger.warning("More than 1 game found, this is currently unsupported")
    return os.path.join(gog_game_path, games[0])


def get_gog_config(gog_game_path):
    """Extract runtime information such as executable paths from GOG files"""
    config_filename = [fn for fn in os.listdir(gog_game_path) if fn.startswith("goggame") and fn.endswith(".info")]
    if not config_filename:
        logger.error("No config file found in %s", gog_game_path)
        return
    gog_config_path = os.path.join(gog_game_path, config_filename[0])
    with open(gog_config_path, encoding="utf-8") as gog_config_file:
        gog_config = json.loads(gog_config_file.read())
    return gog_config


def get_game_config(task, gog_game_path):
    def resolve_path(path):
        """GOG's paths are relative to the gog_game_path, not relative to each other,
        so we resolve them all to absolute paths and fix casing issues. If required,
        we'll replace backslashes with slashes."""
        resolved = system.fix_path_case(os.path.join(gog_game_path, path))
        if os.path.exists(resolved):
            return resolved

        fixed = path.replace("\\", "/")
        if fixed.startswith("/"):
            fixed = fixed[1:]
        resolved = system.fix_path_case(os.path.join(gog_game_path, fixed))
        if os.path.exists(resolved):
            return resolved

        logger.warning("GOG configuration path '%s' could not be resolved", path)
        return path

    config = {}
    if "path" not in task:
        return

    config["exe"] = resolve_path(task["path"])
    if task.get("workingDir"):
        config["working_dir"] = resolve_path(task["workingDir"])
    if task.get("arguments"):
        config["args"] = task["arguments"]
    if task.get("name"):
        config["name"] = task["name"]
    return config


def convert_gog_config_to_lutris(gog_config, gog_game_path):
    play_tasks = gog_config["playTasks"]
    lutris_config = {"launch_configs": []}
    for task in play_tasks:
        config = get_game_config(task, gog_game_path)
        if not config:
            continue
        if task.get("isPrimary"):
            lutris_config.update(config)
        else:
            lutris_config["launch_configs"].append(config)
    return lutris_config


def get_gog_config_from_path(target_path):
    """Return the GOG configuration for a root path"""
    gog_game_path = get_gog_game_path(target_path)
    if gog_game_path:
        return get_gog_config(gog_game_path)



"""Window used for game installers"""

# pylint: disable=too-many-lines
import os
import traceback
from gettext import gettext as _
from typing import List

from gi.repository import Gdk, Gio, GLib, Gtk

from lutris import settings
from lutris.config import LutrisConfig
from lutris.game import Game
from lutris.gui.dialogs import (
    DirectoryDialog,
    InstallerSourceDialog,
    ModelessDialog,
    QuestionDialog,
    display_error,
)
from lutris.gui.dialogs.cache import CacheConfigurationDialog
from lutris.gui.dialogs.delegates import DialogInstallUIDelegate
from lutris.gui.installer.files_box import InstallerFilesBox
from lutris.gui.installer.script_picker import InstallerPicker
from lutris.gui.widgets import NotificationSource
from lutris.gui.widgets.common import FileChooserEntry
from lutris.gui.widgets.log_text_view import LogTextView
from lutris.gui.widgets.navigation_stack import NavigationStack
from lutris.gui.widgets.utils import get_main_window
from lutris.installer import InstallationKind, interpreter
from lutris.installer.errors import MissingGameDependencyError, ScriptingError
from lutris.installer.interpreter import ScriptInterpreter
from lutris.util import xdgshortcuts
from lutris.util.jobs import AsyncCall
from lutris.util.linux import LINUX_SYSTEM
from lutris.util.log import get_log_contents, logger
from lutris.util.steam import shortcut as steam_shortcut
from lutris.util.strings import gtk_safe, human_size
from lutris.util.system import is_removeable

INSTALLATION_FAILED = NotificationSource()
INSTALLATION_COMPLETED = NotificationSource()


class MarkupLabel(Gtk.Label):
    """Label for installer window"""

    def __init__(self, markup=None, **kwargs):
        super().__init__(label=markup, use_markup=True, wrap=True, justify=Gtk.Justification.CENTER, **kwargs)
        self.set_alignment(0.5, 0)


class InstallerWindow(ModelessDialog, DialogInstallUIDelegate, ScriptInterpreter.InterpreterUIDelegate):  # type:ignore[misc]
    """GUI for the install process.

    This window is divided into pages; as you go through the install each page
    is created and displayed to you. You can also go back and visit previous pages
    again. Going *forward* triggers installation work- it does not all way until the
    very end.

    Most pages are defined by a load_X_page() function that initializes the page
    when arriving at it, and which presents it. But this uses a present_X_page() page
    that shows the page (and is used alone for the 'Back' button), and this in turn
    uses a create_X_page() function to create the page the first time it is visited.
    """

    def __init__(self, installers, service=None, appid=None, installation_kind=InstallationKind.INSTALL, **kwargs):
        ModelessDialog.__init__(self, use_header_bar=True, **kwargs)
        ScriptInterpreter.InterpreterUIDelegate.__init__(self, service, appid)
        self.set_default_size(740, 460)
        self.installers = installers
        self.config = {}
        self.selected_extras = []
        self.install_in_progress = False
        self.install_complete = False
        self.interpreter = None
        self.installation_kind = installation_kind
        self.continue_handler = None

        self.accelerators = Gtk.AccelGroup()
        self.add_accel_group(self.accelerators)

        content_area = self.get_content_area()

        content_area.set_margin_top(18)
        content_area.set_margin_bottom(18)
        content_area.set_margin_right(18)
        content_area.set_margin_left(18)
        content_area.set_spacing(12)

        # Header labels
        self.status_label = MarkupLabel(no_show_all=True)
        content_area.pack_start(self.status_label, False, False, 0)

        # Header bar buttons
        self.back_button = self.add_start_button(_("Back"), self.on_back_clicked)
        self.back_button.set_no_show_all(True)
        key, mod = Gtk.accelerator_parse("<Alt>Left")
        self.back_button.add_accelerator("clicked", self.accelerators, key, mod, Gtk.AccelFlags.VISIBLE)
        key, mod = Gtk.accelerator_parse("<Alt>Home")
        self.accelerators.connect(key, mod, Gtk.AccelFlags.VISIBLE, self.on_navigate_home)

        self.cancel_button = self.add_start_button(_("Cancel"), self.on_cancel_clicked)
        self.get_header_bar().set_show_close_button(False)

        self.continue_button = self.add_end_button(_("_Continue"))

        # The cancel button doubles as 'Close' and 'Abort' depending on the state of the install
        key, mod = Gtk.accelerator_parse("Escape")
        self.cancel_button.add_accelerator("clicked", self.accelerators, key, mod, Gtk.AccelFlags.VISIBLE)

        # Navigation stack
        self.stack = NavigationStack(self.back_button, cancel_button=self.cancel_button)
        self.register_page_creators()
        content_area.pack_start(self.stack, True, True, 0)

        # Menu buttons
        menu_icon = Gtk.Image.new_from_icon_name("open-menu-symbolic", Gtk.IconSize.MENU)
        self.menu_button = Gtk.MenuButton(child=menu_icon)
        self.menu_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, visible=True, halign=Gtk.Align.END)
        self.menu_box.set_border_width(9)
        self.menu_box.set_spacing(3)
        self.menu_box.set_can_focus(False)
        self.menu_button.set_popover(Gtk.Popover(child=self.menu_box, can_focus=False, relative_to=self.menu_button))
        self.get_header_bar().pack_end(self.menu_button)

        self.cache_button = self.add_menu_button(
            _("Configure download cache"),
            self.on_cache_clicked,
            tooltip=_("Change where Lutris downloads game installer files."),
        )

        self.source_button = self.add_menu_button(_("View installer source"), self.on_source_clicked)

        # Pre-create some UI bits we need to refer to in several places.
        # (We lazy allocate more of it, but these are a pain.)

        self.extras_tree_store = Gtk.TreeStore(
            bool,  # is selected?
            bool,  # is inconsistent?
            object,  # extras dict
            str,  # label
        )

        self.location_entry = FileChooserEntry(
            "Select folder",
            Gtk.FileChooserAction.SELECT_FOLDER,
            warn_if_non_empty=True,
            warn_if_non_writable_parent=True,
            warn_if_ntfs=True,
        )
        self.location_entry.connect("changed", self.on_location_entry_changed)

        self.installer_files_box = InstallerFilesBox()
        self.installer_files_box.connect("files-available", self.on_files_available)
        self.installer_files_box.connect("files-ready", self.on_files_ready)

        self.log_buffer = Gtk.TextBuffer()
        self.error_details_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6, no_show_all=True)
        self.error_details_buffer = Gtk.TextBuffer()
        self.error_reporter = self.load_error_page

        # And... go!
        self.load_first_page()
        self.show_all()
        self.present()

    def add_start_button(self, label, handler=None, tooltip=None, sensitive=True):
        button = Gtk.Button.new_with_mnemonic(label)
        button.set_sensitive(sensitive)
        button.set_no_show_all(True)
        if tooltip:
            button.set_tooltip_text(tooltip)
        if handler:
            button.connect("clicked", handler)

        header_bar = self.get_header_bar()
        header_bar.pack_start(button)
        return button

    def add_end_button(self, label, handler=None, tooltip=None, sensitive=True):
        """Add a button to the action buttons box"""
        button = Gtk.Button.new_with_mnemonic(label)
        button.set_sensitive(sensitive)
        button.set_no_show_all(True)
        if tooltip:
            button.set_tooltip_text(tooltip)
        if handler:
            button.connect("clicked", handler)

        header_bar = self.get_header_bar()
        header_bar.pack_end(button)
        return button

    def add_menu_button(self, label, handler=None, tooltip=None, sensitive=True):
        """Add a button to the menu in the header bar"""
        button = Gtk.ModelButton(label, visible=True, xalign=0.0)
        button.set_sensitive(sensitive)
        button.set_no_show_all(True)
        if tooltip:
            button.set_tooltip_text(tooltip)
        if handler:
            button.connect("clicked", handler)

        self.menu_box.pack_start(button, False, False, 0)
        return button

    def on_cache_clicked(self, _button):
        """Open the cache configuration dialog"""
        CacheConfigurationDialog(parent=self)

    def on_response(self, dialog, response: Gtk.ResponseType) -> None:
        if response in (Gtk.ResponseType.CLOSE, Gtk.ResponseType.CANCEL, Gtk.ResponseType.DELETE_EVENT):
            self.on_cancel_clicked()
        else:
            super().on_response(dialog, response)

    def on_back_clicked(self, _button):
        self.stack.navigate_back()

    def on_navigate_home(self, _accel_group, _window, _keyval, _modifier):
        self.stack.navigate_home()

    def on_cancel_clicked(self, _button=None):
        """Ask a confirmation before cancelling the installation, if it has started."""

        def on_cancelled():
            if self.interpreter:
                self.interpreter.cleanup()  # still remove temporary downloads in any case

            if self.interpreter and not self.install_in_progress:
                INSTALLATION_COMPLETED.fire()
            else:
                INSTALLATION_FAILED.fire()

            self.destroy()

        if self.install_in_progress:
            widgets = []

            remove_checkbox = Gtk.CheckButton.new_with_label(_("Remove game files"))
            if (
                self.interpreter
                and self.interpreter.target_path
                and self.interpreter.game_dir_created
                and self.installation_kind == InstallationKind.INSTALL
                and is_removeable(self.interpreter.target_path, LutrisConfig().system_config)
            ):
                remove_checkbox.set_active(self.interpreter.game_dir_created)
                remove_checkbox.show()
                widgets.append(remove_checkbox)

            confirm_cancel_dialog = QuestionDialog(
                {
                    "parent": self,
                    "question": _("Are you sure you want to cancel the installation?"),
                    "title": _("Cancel installation?"),
                    "widgets": widgets,
                }
            )
            if confirm_cancel_dialog.result != Gtk.ResponseType.YES:
                logger.debug("User aborted installation cancellation")
                return

            self.installer_files_box.stop_all()
            if self.interpreter:
                self.interpreter.revert(
                    remove_game_dir=remove_checkbox.get_active(),
                    completion_function=on_cancelled,
                    error_function=self.on_signal_error,
                )
            else:
                on_cancelled()
        else:
            self.installer_files_box.stop_all()
            on_cancelled()

    def on_source_clicked(self, _button):
        InstallerSourceDialog(self.interpreter.installer.script_pretty, self.interpreter.installer.game_name, self)

    def on_signal_error(self, error):
        self._handle_callback_error(error)

    def on_idle_error(self, error):
        self._handle_callback_error(error)

    def _handle_callback_error(self, error):
        if self.install_in_progress:
            self.load_error_page(error)
        else:
            display_error(error, parent=self)
            self.stack.navigation_reset()

    def set_status(self, markup):
        """Display a short status text."""
        self.status_label.set_markup(markup)
        self.status_label.set_visible(bool(markup))

    def get_status(self):
        return self.status_label.get_text() if self.status_label.get_visible() else ""

    def register_page_creators(self):
        self.stack.add_named_factory("choose_installer", self.create_choose_installer_page)
        self.stack.add_named_factory("destination", self.create_destination_page)
        self.stack.add_named_factory("installer_files", self.create_installer_files_page)
        self.stack.add_named_factory("extras", self.create_extras_page)
        self.stack.add_named_factory("spinner", self.create_spinner_page)
        self.stack.add_named_factory("log", self.create_log_page)
        self.stack.add_named_factory("error", self.create_error_page)
        self.stack.add_named_factory("nothing", lambda *x: Gtk.Box())

    def load_first_page(self) -> None:
        # If we're downloading updates in the background, we'll
        # put up a spinner page to wait until that's done. Installations can
        # fail if Lutris components are missing, and users sometimes try to install
        # a game just after their first Lutris startup. This should help.
        window = get_main_window()
        if window and not window.download_queue.is_empty:
            download_queue = window.download_queue

            def on_start_installation(*args):
                self.load_choose_installer_page()
                download_queue.disconnect(dc_handler)

            def on_download_complete(*args):
                if download_queue.is_empty:
                    on_start_installation()

            dc_handler = download_queue.connect("download-completed", on_download_complete)
            self.load_spinner_page(
                _(
                    "Waiting for Lutris component installation\n"
                    "<small>Installations can fail if Lutris components are not installed first.</small>"
                )
            )
            self.display_continue_button(on_start_installation)
        else:
            self.load_choose_installer_page()

    # Interpreter UI Delegate
    #
    # These methods are called from the ScriptInterpreter, and defer work until idle time
    # so the installation itself is not interrupted or paused for UI updates.

    def report_error(self, error):
        GLib.idle_add(self.error_reporter, error)

    def report_status(self, status):
        GLib.idle_add(self.set_status, gtk_safe(status))

    def attach_log(self, command):
        # Hook the log buffer right now, lest we miss updates.
        command.set_log_buffer(self.log_buffer)
        GLib.idle_add(self.load_log_page)

    def begin_disc_prompt(self, message_markup, requires, installer, callback):
        GLib.idle_add(
            self.load_ask_for_disc_page,
            message_markup,
            requires,
            installer,
            callback,
        )

    def begin_input_menu(self, alias, options, preselect, callback):
        GLib.idle_add(self.load_input_menu_page, alias, options, preselect, callback)

    def report_finished(self, game_id, status):
        GLib.idle_add(self.load_finish_install_page, game_id, gtk_safe(status))

    # Choose Installer Page
    #
    # This page offers a choice of installer scripts to run.

    def load_choose_installer_page(self):
        self.validate_scripts(self.installers)
        self.stack.navigate_to_page(self.present_choose_installer_page)

    def create_choose_installer_page(self):
        installer_picker = InstallerPicker(self.installers)
        installer_picker.connect("installer-selected", self.on_installer_selected)
        return Gtk.ScrolledWindow(
            hexpand=True, vexpand=True, child=installer_picker, shadow_type=Gtk.ShadowType.ETCHED_IN
        )

    def present_choose_installer_page(self):
        """Stage where we choose an install script."""
        self.set_status("")
        self.set_title(_("Install %s") % self.installers[0]["name"])
        self.stack.present_page("choose_installer")
        self.display_cancel_button(extra_buttons=[self.cache_button])

    def on_installer_selected(self, _widget, installer_version):
        """Sets the script interpreter to the correct script then proceed to
        install folder selection.

        If the installed game depends on another one and it's not installed,
        prompt the user to install it and quit this installer.
        """
        try:
            script = None
            for _script in self.installers:
                if _script["version"] == installer_version:
                    script = _script
            self.interpreter = interpreter.ScriptInterpreter(script, self)
            self.interpreter.connect("runners-installed", self.on_runners_ready)
        except MissingGameDependencyError as ex:
            dlg = QuestionDialog(
                {
                    "parent": self,
                    "question": _("This game requires %s. Do you want to install it?") % ex.slug,
                    "title": _("Missing dependency"),
                }
            )
            if dlg.result == Gtk.ResponseType.YES:
                application = Gio.Application.get_default()
                application.show_lutris_installer_window(game_slug=ex.slug)
            return

        self.set_title(_("Installing {}").format(self.interpreter.installer.game_name))
        self.load_destination_page()

    def validate_scripts(self, installers):
        """Auto-fixes some script aspects and checks for mandatory fields"""
        if not installers:
            raise ScriptingError(_("No installer available"))
        for script in installers:
            for item in ["description", "notes"]:
                script[item] = script.get(item) or ""
            for item in ["name", "runner", "version"]:
                if item not in script:
                    raise ScriptingError(_('Missing field "%s" in install script') % item)
            for file_desc in script["script"].get("files", {}):
                if len(file_desc) > 1:
                    raise ScriptingError(_('Improperly formatted file "%s"') % file_desc)

    # Destination Page
    #
    # This page selects the directory where the game will be installed,
    # as well as few other minor options.

    def load_destination_page(self):
        """Stage where we select the install directory."""
        if not self.interpreter.installer.creates_game_folder:
            self.on_destination_confirmed()
            return

        default_path = self.interpreter.get_default_target()
        self.location_entry.set_text(default_path)

        self.stack.navigate_to_page(self.present_destination_page)
        self.continue_button.grab_focus()

    def create_destination_page(self):
        installer_create_desktop_shortcut = settings.read_bool_setting("installer_create_desktop_shortcut", False)
        installer_create_menu_shortcut = settings.read_bool_setting("installer_create_menu_shortcut", False)

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)
        vbox.pack_start(self.location_entry, False, False, 0)

        desktop_shortcut_button = Gtk.CheckButton(_("Create desktop shortcut"), visible=True)
        desktop_shortcut_button.set_active(installer_create_desktop_shortcut)
        desktop_shortcut_button.connect("clicked", self.on_create_desktop_shortcut_clicked)
        self.config["create_desktop_shortcut"] = installer_create_desktop_shortcut

        vbox.pack_start(desktop_shortcut_button, False, False, 0)

        menu_shortcut_button = Gtk.CheckButton(_("Create application menu shortcut"), visible=True)
        menu_shortcut_button.set_active(installer_create_menu_shortcut)
        menu_shortcut_button.connect("clicked", self.on_create_menu_shortcut_clicked)
        self.config["create_menu_shortcut"] = installer_create_menu_shortcut

        vbox.pack_start(menu_shortcut_button, False, False, 0)

        if steam_shortcut.vdf_file_exists():
            steam_shortcut_button = Gtk.CheckButton(_("Create Steam shortcut"), visible=True)
            steam_shortcut_button.set_active(settings.read_bool_setting("installer_create_steam_shortcut", False))
            steam_shortcut_button.connect("clicked", self.on_create_steam_shortcut_clicked)
            vbox.pack_start(steam_shortcut_button, False, False, 0)
        return vbox

    def present_destination_page(self):
        """Display the destination chooser."""

        self.set_status(_("Select installation directory"))
        self.stack.present_page("destination")
        self.display_continue_button(
            self.on_destination_confirmed, extra_buttons=[self.cache_button, self.source_button]
        )

    def on_destination_confirmed(self, _button=None):
        """Let the interpreter take charge of the next stages."""

        self.load_spinner_page(
            _("Preparing Lutris for installation"),
            cancellable=False,
            extra_buttons=[self.cache_button, self.source_button],
        )
        GLib.idle_add(self.launch_install)

    def launch_install(self):
        if not self.interpreter.launch_install(self):
            self.stack.navigation_reset()

    def on_location_entry_changed(self, entry, _data=None):
        """Set the installation target for the game."""
        self.interpreter.target_path = os.path.expanduser(entry.get_text())

    def on_create_desktop_shortcut_clicked(self, checkbutton):
        settings.write_setting("installer_create_desktop_shortcut", checkbutton.get_active())
        self.config["create_desktop_shortcut"] = checkbutton.get_active()

    def on_create_menu_shortcut_clicked(self, checkbutton):
        settings.write_setting("installer_create_menu_shortcut", checkbutton.get_active())
        self.config["create_menu_shortcut"] = checkbutton.get_active()

    def on_create_steam_shortcut_clicked(self, checkbutton):
        settings.write_setting("installer_create_steam_shortcut", checkbutton.get_active())
        self.config["create_steam_shortcut"] = checkbutton.get_active()

    def on_runners_ready(self, _widget=None):
        AsyncCall(self.interpreter.get_extras, self.on_extras_loaded)

    # Extras Page
    #
    # This pages offers to download the extras that come with the game; the
    # user specifies the specific extras desired.
    #
    # If there are no extras, the page triggers as if the user had clicked 'Continue',
    # moving on to pre-installation, then the installer files page.

    def get_extra_label(self, extra):
        """Return a label for the extras picker"""
        label = extra["name"]
        _infos = []
        if extra.get("total_size"):
            _infos.append(human_size(extra["total_size"]))
        if extra.get("type"):
            _infos.append(extra["type"])
        if _infos:
            label += " (%s)" % ", ".join(_infos)
        return label

    def on_extras_loaded(self, all_extras, error):
        if error:
            self._handle_callback_error(error)
            return

        if all_extras:
            self.extras_tree_store.clear()
            for extra_source, extras in all_extras.items():
                parent = self.extras_tree_store.append(None, (None, None, None, extra_source))
                for extra in extras:
                    self.extras_tree_store.append(parent, (False, False, extra, self.get_extra_label(extra)))

            self.stack.navigate_to_page(self.present_extras_page)
        else:
            self.on_extras_ready()

    def create_extras_page(self):
        treeview = Gtk.TreeView(self.extras_tree_store)
        treeview.set_headers_visible(False)
        treeview.expand_all()
        renderer_toggle = Gtk.CellRendererToggle()
        renderer_toggle.connect("toggled", self.on_extra_toggled, self.extras_tree_store)
        renderer_text = Gtk.CellRendererText()

        installed_column = Gtk.TreeViewColumn(None, renderer_toggle, active=0, inconsistent=1)
        treeview.append_column(installed_column)

        label_column = Gtk.TreeViewColumn(None, renderer_text)
        label_column.add_attribute(renderer_text, "text", 3)
        label_column.set_property("min-width", 80)
        treeview.append_column(label_column)

        return Gtk.ScrolledWindow(
            hexpand=True, vexpand=True, child=treeview, visible=True, shadow_type=Gtk.ShadowType.ETCHED_IN
        )

    def present_extras_page(self):
        """Show installer screen with the extras picker"""
        logger.debug("Showing extras page")

        def on_continue(_button):
            self.on_extras_confirmed(self.extras_tree_store)

        self.set_status(
            _(
                "This game has extra content\n"
                "<small>Select which one you want and "
                "they will be available in the 'extras' folder where the game is installed.</small>"
            )
        )
        self.stack.present_page("extras")
        self.display_continue_button(on_continue, extra_buttons=[self.cache_button, self.source_button])

    def on_extra_toggled(self, _widget, path, model):
        toggled_row = model[path]
        toggled_row_iter = model.get_iter(path)

        toggled_row[0] = not toggled_row[0]
        toggled_row[1] = False

        if model.iter_has_child(toggled_row_iter):
            extra_iter = model.iter_children(toggled_row_iter)
            while extra_iter:
                extra_row = model[extra_iter]
                extra_row[0] = toggled_row[0]
                extra_iter = model.iter_next(extra_iter)
        else:
            for heading_row in model:
                all_extras_active = True
                any_extras_active = False
                extra_iter = model.iter_children(heading_row.iter)
                while extra_iter:
                    extra_row = model[extra_iter]
                    if extra_row[0]:
                        any_extras_active = True
                    else:
                        all_extras_active = False
                    extra_iter = model.iter_next(extra_iter)

                heading_row[0] = all_extras_active
                heading_row[1] = any_extras_active

    def on_extras_confirmed(self, extra_store):
        """Resume install when user has selected extras to download"""
        selected_extras = []

        def save_extra(model, path, iter_):
            selected, _inconsistent, extra, _label = model[iter_]
            if selected and extra:
                selected_extras.append(extra)

        extra_store.foreach(save_extra)

        self.selected_extras = selected_extras
        GLib.idle_add(self.on_extras_ready)

    def on_extras_ready(self, *args):
        self.load_installer_files_page()

    # Installer Files & Downloading Page
    #
    # This page shows the files that are needed, and can download them. The user can
    # also select pre-existing files. The downloading page uses the same page widget,
    # but different buttons at the bottom.

    def load_installer_files_page(self):
        if self.installation_kind == InstallationKind.UPDATE:
            patch_version = self.interpreter.installer.version
        else:
            patch_version = None
        self.load_spinner_page(_("Preparing game files..."), cancellable=False)
        AsyncCall(
            self.interpreter.installer.prepare_game_files, self.on_files_prepared, self.selected_extras, patch_version
        )

    def on_files_prepared(self, _result, error):
        if error:
            self._handle_callback_error(error)
            return

        if not self.interpreter.installer.files:
            logger.debug("Installer doesn't require files")
            self.launch_installer_commands()
            return

        logger.debug("Game files prepared.")
        self.installer_files_box.load_installer(self.interpreter.installer)
        self.stack.navigate_to_page(self.present_installer_files_page)

    def create_installer_files_page(self):
        return Gtk.ScrolledWindow(
            hexpand=True,
            vexpand=True,
            child=self.installer_files_box,
            visible=True,
            shadow_type=Gtk.ShadowType.ETCHED_IN,
        )

    def present_installer_files_page(self):
        """Show installer screen with the file picker / downloader"""
        logger.debug("Presenting installer files page")
        self.set_status(_("Please review the files needed for the installation then click 'Install'"))
        self.stack.present_page("installer_files")
        self.display_install_button(self.on_files_confirmed, sensitive=self.installer_files_box.is_ready)

    def present_downloading_files_page(self):
        def on_exit_page():
            self.installer_files_box.stop_all()

        self.set_status(_("Downloading game data"))
        self.stack.present_page("installer_files")
        self.display_install_button(None, sensitive=False)
        return on_exit_page

    def on_files_ready(self, _widget, files_ready):
        """Toggle state of continue button based on ready state"""
        self.display_install_button(self.on_files_confirmed, sensitive=files_ready)

    def on_files_confirmed(self, _button):
        """Call this when the user confirms the install files
        This will start the downloads.
        """
        try:
            self.installer_files_box.start_all()
            self.stack.jump_to_page(self.present_downloading_files_page)
        except PermissionError as ex:
            raise ScriptingError(_("Unable to get files: %s") % ex) from ex

    def on_files_available(self, widget):
        """All files are available, continue the install"""
        logger.info("All files are available, continuing install")
        self.interpreter.game_files = widget.get_game_files()
        # Idle-add here to ensure that the launch occurs after
        # on_files_confirmed(), since they can race when no actual
        # download is required.
        GLib.idle_add(self.launch_installer_commands)

    def launch_installer_commands(self):
        logger.info("Launching installer commands")
        self.install_in_progress = True
        self.load_spinner_page(_("Installing game data"))
        self.stack.discard_navigation()  # once we really start installing, no going back!
        self.interpreter.installer.install_extras()
        self.interpreter.launch_installer_commands()

    # Spinner Page
    #
    # Provides a generic progress spinner and displays a status. The back button
    # is disabled for this page.

    def load_spinner_page(self, status: str, cancellable: bool = True, extra_buttons: List[Gtk.Button] = None) -> None:
        def present_spinner_page():
            """Show a spinner in the middle of the view"""

            def on_exit_page():
                self.stack.set_back_allowed(True)

            self.set_status(status)
            self.stack.present_page("spinner")

            if cancellable:
                self.display_cancel_button(extra_buttons=extra_buttons)
            else:
                self.display_buttons(extra_buttons or [], False)

            self.stack.set_back_allowed(False)
            return on_exit_page

        self.stack.jump_to_page(present_spinner_page)

    def create_spinner_page(self):
        spinner = Gtk.Spinner(halign=Gtk.Align.CENTER, valign=Gtk.Align.CENTER)
        spinner.start()
        return spinner

    # Log Page
    #
    # This page shos a LogTextView where an installer command can display
    # output. This appears when summons by the installer script.

    def load_log_page(self):
        self.stack.jump_to_page(self.present_log_page)

    def create_log_page(self):
        log_textview = LogTextView(self.log_buffer)
        return Gtk.ScrolledWindow(hexpand=True, vexpand=True, child=log_textview, shadow_type=Gtk.ShadowType.ETCHED_IN)

    def present_log_page(self):
        """Creates a TextBuffer and attach it to a command"""

        def on_exit_page():
            self.error_reporter = saved_reporter

        def on_error(error):
            self.set_status(gtk_safe(error))

        saved_reporter = self.error_reporter
        self.error_reporter = on_error
        self.stack.present_page("log")
        self.display_cancel_button()
        return on_exit_page

    # Input Menu Page
    #
    # This page shows a list of choices to the user, and calls
    # back into a callback when the user makes a choice. This is summoned
    # by the installer script as well.

    def load_input_menu_page(self, alias, options, preselect, callback):
        def present_input_menu_page():
            """Display an input request as a dropdown menu with options."""

            def on_continue(_button):
                try:
                    callback(alias, combobox)
                    self.stack.restore_current_page(previous_page)
                except Exception as err:
                    # If the callback fails, the installation does not continue
                    # to run, so we'll go to error page.
                    self.load_error_page(err)

            model = Gtk.ListStore(str, str)

            for option in options:
                key, label = option.popitem()
                model.append([key, label])

            combobox = Gtk.ComboBox.new_with_model(model)
            renderer_text = Gtk.CellRendererText()
            combobox.pack_start(renderer_text, True)
            combobox.add_attribute(renderer_text, "text", 1)
            combobox.set_id_column(0)
            combobox.set_halign(Gtk.Align.CENTER)
            combobox.set_valign(Gtk.Align.START)
            combobox.set_active_id(preselect)
            combobox.connect("changed", self.on_input_menu_changed)

            self.stack.present_replacement_page("input_menu", combobox)
            self.display_continue_button(on_continue)
            self.continue_button.grab_focus()
            self.on_input_menu_changed(combobox)

        # we must use jump_to_page() here since it would be unsave to return
        # back to this page and re-execute the callback.
        previous_page = self.stack.save_current_page()
        self.stack.jump_to_page(present_input_menu_page)

    def on_input_menu_changed(self, combobox):
        """Enable continue button if a non-empty choice is selected"""
        self.continue_button.set_sensitive(bool(combobox.get_active_id()))

    # Ask for Disc Page
    #
    # This page asks the user for a disc; it also has a callback used when
    # the user selects a disc. Again, this is summoned by the installer script.

    def load_ask_for_disc_page(self, message_markup, requires, installer, callback):
        def present_ask_for_disc_page():
            """Ask the user to do insert a CD-ROM."""

            def wrapped_callback(*args, **kwargs):
                try:
                    callback(*args, **kwargs)
                    self.stack.restore_current_page(previous_page)
                except Exception as err:
                    # If the callback fails, the installation does not continue
                    # to run, so we'll go to error page.
                    self.load_error_page(err)

            vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)
            label = MarkupLabel(message_markup)
            vbox.pack_start(label, False, False, 0)

            buttons_box = Gtk.Box()
            buttons_box.set_margin_top(40)
            buttons_box.set_margin_bottom(40)
            vbox.pack_start(buttons_box, False, False, 0)

            if not LINUX_SYSTEM.is_flatpak():
                # Lutris flatplak doesn't autodetect files on CD-ROM properly
                # and selecting this option doesn't let the user click "Back"
                # so the only option is to cancel the install.
                autodetect_button = Gtk.Button(label=_("Autodetect"))
                autodetect_button.connect("clicked", wrapped_callback, requires)
                autodetect_button.grab_focus()
                buttons_box.pack_start(autodetect_button, True, True, 40)

            browse_button = Gtk.Button(label=_("Browse…"))
            callback_data = {"callback": wrapped_callback, "requires": requires}
            browse_button.connect("clicked", self.on_browse_clicked, callback_data)
            buttons_box.pack_start(browse_button, True, True, 40)

            self.stack.present_replacement_page("ask_for_disc", vbox)
            if installer.runner == "wine":
                eject_button = Gtk.Button(_("Eject"), halign=Gtk.Align.END)
                eject_button.connect("clicked", self.on_eject_clicked)
                vbox.pack_end(eject_button, False, False, 0)
                vbox.pack_end(Gtk.Separator(), False, False, 0)

            vbox.show_all()
            self.display_cancel_button()

        previous_page = self.stack.save_current_page()
        self.stack.jump_to_page(present_ask_for_disc_page)

    def on_browse_clicked(self, widget, callback_data):
        dialog = DirectoryDialog(_("Select the folder where the disc is mounted"), parent=self)
        folder = dialog.folder
        callback = callback_data["callback"]
        requires = callback_data["requires"]
        callback(widget, requires, folder)

    def on_eject_clicked(self, _widget, data=None):
        self.interpreter.eject_wine_disc()

    # Error Message Page
    #
    # This is used to display an error; such an error halts the installation,
    # and isn't recoverable. Used by the installer script.

    def load_error_page(self, error: BaseException) -> None:
        self.stack.navigate_to_page(lambda *x: self.present_error_page(error))
        self.stack.set_back_allowed(False)
        self.cancel_button.grab_focus()

    def present_error_page(self, error: BaseException) -> None:
        self.set_status(gtk_safe(str(error)))

        is_expected = hasattr(error, "is_expected") and error.is_expected

        if is_expected:
            formatted = traceback.format_exception(type(error), error, error.__traceback__)
            formatted = "\n".join(formatted).strip()

            log = get_log_contents()
            if log:
                formatted = f"{formatted}\n\nLutris log:\n{log}".strip()

            self.error_details_buffer.set_text(formatted)

        self.error_details_box.set_visible(not is_expected)

        self.stack.present_page("error")
        self.display_cancel_button()

    def create_error_page(self) -> Gtk.Widget:
        def on_copy_clicked(_button):
            status = self.get_status()
            details = self.error_details_buffer.get_text(
                self.error_details_buffer.get_start_iter(), self.error_details_buffer.get_end_iter(), True
            )
            text = f"{status}\n\n{details}"
            clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
            clipboard.set_text(text, -1)

        error_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)

        label = Gtk.Label(xalign=0.0, wrap=True)
        label.set_markup(
            _(
                "An unexpected error has occurred while installing this game. "
                "Please share the details below with the Lutris team on "
                "<a href='https://github.com/lutris/lutris'>GitHub</a> or "
                "<a href='https://discordapp.com/invite/Pnt5CuY'>Discord</a>."
            )
        )
        self.error_details_box.pack_start(label, False, False, 0)

        frame = Gtk.Frame(shadow_type=Gtk.ShadowType.ETCHED_IN)

        details_textview = Gtk.TextView(editable=False, buffer=self.error_details_buffer)

        scrolledwindow = Gtk.ScrolledWindow()
        scrolledwindow.add(details_textview)
        frame.add(scrolledwindow)
        self.error_details_box.pack_start(frame, True, True, 0)
        error_box.pack_start(self.error_details_box, True, True, 0)

        copy_button = Gtk.Button(label=_("Copy Details to Clipboard"), halign=Gtk.Align.START)
        error_box.pack_end(copy_button, False, True, 0)
        copy_button.connect("clicked", on_copy_clicked)

        return error_box

    # Finished Page
    #
    # This is used to inidcate that the install is complete. The user
    # can launch the game a this point, or just close out of the window.
    #
    # Loading this page does some final installation steps before the UI updates.

    def load_finish_install_page(self, game_id, status):
        if self.config.get("create_desktop_shortcut"):
            AsyncCall(self.create_shortcut, None, True)
        if self.config.get("create_menu_shortcut"):
            AsyncCall(self.create_shortcut, None)

        # Save game to trigger a game-updated signal,
        # but take care not to create a blank game
        if game_id:
            game = Game(game_id)
            if self.config.get("create_steam_shortcut"):
                AsyncCall(steam_shortcut.create_shortcut, None, game)
            game.save()

        self.install_in_progress = False
        self.install_complete = True

        self.stack.jump_to_page(lambda *x: self.present_finished_page(game_id, status))
        self.stack.discard_navigation()
        self.cancel_button.grab_focus()

        if not self.is_active():
            self.set_urgency_hint(True)  # Blink in taskbar
            self.connect("focus-in-event", self.on_window_focus)

    def present_finished_page(self, game_id, status):
        self.set_status(status)
        self.stack.present_page("nothing")
        self.display_continue_button(self.on_launch_clicked, continue_button_label=_("_Launch"), suggested_action=False)

    def on_launch_clicked(self, button):
        """Launch a game after it's been installed."""
        button.set_sensitive(False)
        game = Game(self.interpreter.installer.game_id)
        if game.is_db_stored:
            # Since we're closing this window, we can't use
            # it as the delegate.
            application = Gio.Application.get_default()
            game.launch(application.launch_ui_delegate)
        else:
            logger.error("Game has no ID, launch button should not be drawn")
        self.on_cancel_clicked(button)

    def on_window_focus(self, _widget, *_args):
        """Remove urgency hint (flashing indicator) when window receives focus"""
        self.set_urgency_hint(False)

    def create_shortcut(self, desktop=False):
        """Create desktop or global menu shortcuts."""
        game_slug = self.interpreter.installer.game_slug
        game_id = self.interpreter.installer.game_id
        game_name = self.interpreter.installer.game_name

        if desktop:
            xdgshortcuts.create_launcher(game_slug, game_id, game_name, desktop=True)
        else:
            xdgshortcuts.create_launcher(game_slug, game_id, game_name, menu=True)

    # Buttons
    def display_continue_button(
        self, handler, continue_button_label=_("_Continue"), sensitive=True, suggested_action=True, extra_buttons=None
    ):
        """This shows the continue button, the close button, and any extra buttons you
        indicate. This will also set the label and sensitivity of the continue button.

        Finally, you cna provide the clicked handler for the continue button,
        though that can be None to leave it disconnected.

        We call this repeatedly, as we arrive at each page. Each call disconnects
        the previous clicked handler and connects the new one.
        """
        self.continue_button.set_label(continue_button_label)
        self.continue_button.set_sensitive(sensitive)

        style_context = self.continue_button.get_style_context()

        if suggested_action:
            style_context.add_class("suggested-action")
        else:
            style_context.remove_class("suggested-action")

        if self.continue_handler:
            self.continue_button.disconnect(self.continue_handler)

        if handler:
            self.continue_handler = self.continue_button.connect("clicked", handler)
        else:
            self.continue_handler = None

        buttons = [self.continue_button] + (extra_buttons or [])
        self.display_buttons(buttons)

    def display_install_button(self, handler, sensitive=True):
        """Displays the continue button, but labels it 'Install'."""
        self.display_continue_button(
            handler, continue_button_label=_("_Install"), sensitive=sensitive, extra_buttons=[self.source_button]
        )

    def display_cancel_button(self, extra_buttons=None):
        self.display_buttons(extra_buttons or [], True)

    def display_buttons(self, buttons, cancel_sensitive=True):
        """Shows exactly the buttons given, and hides the others. Updates the close button
        according to whether the install has started."""

        style_context = self.cancel_button.get_style_context()

        if self.install_in_progress:
            self.cancel_button.set_label(_("_Abort"))
            self.cancel_button.set_tooltip_text(_("Abort and revert the installation"))
            style_context.add_class("destructive-action")
        else:
            self.cancel_button.set_label(_("_Close") if self.install_complete else _("Cancel"))
            self.cancel_button.set_tooltip_text("")
            style_context.remove_class("destructive-action")

        self.cancel_button.set_sensitive(cancel_sensitive)

        all_buttons = [self.cache_button, self.source_button, self.continue_button]

        for b in all_buttons:
            b.set_visible(b in buttons)

        any_visible = False
        for b in self.menu_box.get_children():
            if b.get_visible():
                any_visible = True
                break
        self.menu_button.set_visible(any_visible)



"""Window used for game installers"""

# pylint: disable=too-many-lines
from gi.repository import Gtk


class NavigationStack(Gtk.Stack):
    """
    This is a Stack widget that supports a back button and
    lazy-creation of pages.

    Pages should be set up via add_named_factory(), then displayed
    with present_page().

    However, you are meant to have 'present_X_page' functions
    that you pass to navigate_to_page(); this tracks the pages
    you visit, and when you navigate back,the presenter function
    will be called again.

    A presenter function can do more than just call present_page();
    it can configure other aspects of the InstallerWindow. Packaging
    all this into a presenter function keeps things in sync as you navigate.

    A presenter function can return an exit function, called when you navigate away
    from the page again.
    """

    def __init__(self, back_button, cancel_button=None, **kwargs):
        super().__init__(**kwargs)

        self.back_button = back_button
        self.cancel_button = cancel_button
        self.page_factories = {}
        self.stack_pages = {}
        self.navigation_stack = []
        self.navigation_exit_handler = None
        self.current_page_presenter = None
        self.current_navigated_page_presenter = None
        self.back_allowed = True
        self.cancel_allowed = True

    def add_named_factory(self, name, factory):
        """This specifies the factory functioin for the page named;
        this function takes no arguments, but returns the page's widget."""
        self.page_factories[name] = factory

    def set_back_allowed(self, is_allowed=True):
        """This turns the back button off, or back on."""
        self.back_allowed = is_allowed
        self._update_back_button()

    def set_cancel_allowed(self, is_allowed=True):
        """This turns the back button off, or back on."""
        self.cancel_allowed = is_allowed
        self._update_back_button()

    def _update_back_button(self):
        can_go_back = self.back_allowed and self.navigation_stack
        self.back_button.set_visible(can_go_back)
        self.cancel_button.set_visible(not can_go_back and self.cancel_allowed)

    def navigate_to_page(self, page_presenter):
        """Navigates to a page, by invoking 'page_presenter'.

        In addition, this updates the navigation state so navigate_back()
        and such work, they may call the presenter again.
        """
        if self.current_navigated_page_presenter and self.current_navigated_page_presenter != page_presenter:
            self.navigation_stack.append(self.current_navigated_page_presenter)
            self._update_back_button()

        self._go_to_page(page_presenter, True, Gtk.StackTransitionType.SLIDE_LEFT)

    def jump_to_page(self, page_presenter):
        """Jumps to a page, without updating navigation state.

        This does not disturb the behavior of navigate_back().
        This does invoke the exit handler of the current page.
        """
        self._go_to_page(page_presenter, False, Gtk.StackTransitionType.NONE)

    def navigate_back(self):
        """This navigates to the previous page, if any. This will invoke the
        current page's exit function, and the previous page's presenter function.
        """
        if self.navigation_stack and self.back_allowed:
            try:
                back_to = self.navigation_stack.pop()
                self._go_to_page(back_to, True, Gtk.StackTransitionType.SLIDE_RIGHT)
            finally:
                self._update_back_button()

    def navigate_home(self):
        """This navigates to the first page, effectively navigating back until it
        can go no further back. It does not actually traverse the intermediate pages
        though, but goes directly to the first."""
        if self.navigation_stack and self.back_allowed:
            try:
                home = self.navigation_stack[0]
                self.navigation_stack.clear()
                self._go_to_page(home, True, Gtk.StackTransitionType.SLIDE_RIGHT)
            finally:
                self._update_back_button()

    def navigation_reset(self):
        """This reverse the effect of jump_to_page(), returning to the last
        page actually navigate to."""
        if self.current_navigated_page_presenter:
            if self.current_page_presenter != self.current_navigated_page_presenter:
                self._go_to_page(self.current_navigated_page_presenter, True, Gtk.StackTransitionType.SLIDE_RIGHT)

    def save_current_page(self):
        """Returns a tuple containing information about the current page,
        to pass to restore_current_page()."""
        return (self.current_page_presenter, self.current_navigated_page_presenter)

    def restore_current_page(self, state):
        """Restores the current page to the one in effect when the state was generated.
        This does not disturb the navigation stack."""
        page_presenter, navigated_presenter = state
        navigated = page_presenter == navigated_presenter
        self._go_to_page(page_presenter, navigated, Gtk.StackTransitionType.NONE)

    def _go_to_page(self, page_presenter, navigated, transition_type):
        """Switches to a page. If 'navigated' is True, then when you navigate
        away from this page, it can go on the navigation stack. It should be
        False for 'temporary' pages that are not part of normal navigation."""
        exit_handler = self.navigation_exit_handler
        self.set_transition_type(transition_type)
        self.navigation_exit_handler = page_presenter()
        self.current_page_presenter = page_presenter
        if navigated:
            self.current_navigated_page_presenter = page_presenter
        if exit_handler:
            exit_handler()
        self._update_back_button()

    def discard_navigation(self):
        """This throws away the navigation history, so the back
        button is disabled. Previous pages before the current become
        inaccessible."""
        self.navigation_stack.clear()
        self._update_back_button()

    def present_page(self, name):
        """This displays the page names, creating it if required. It
        also calls show_all() on newly created pages.

        This should be called by your presenter functions."""
        if name not in self.stack_pages:
            factory = self.page_factories[name]
            page = factory()
            page.show_all()

            self.add_named(page, name)
            self.stack_pages[name] = page

        self.set_visible_child_name(name)
        return self.stack_pages[name]

    def present_replacement_page(self, name, page):
        """This displays a page that is given, rather than lazy-creating one. It
        still needs a name, but if you re-use a name this will replace the old page.

        This is useful for pages that need special initialization each time they
        appear, but generally such pages can't be returned to via the back
        button. The caller must protect against this if required.
        """
        old_page = self.stack_pages.get(name)

        if old_page != page:
            if old_page:
                self.remove(old_page)

            page.show_all()

            self.add_named(page, name)
            self.stack_pages[name] = page

        self.set_visible_child_name(name)
        return page



"""Installer specific exceptions"""

import sys
from gettext import gettext as _

from lutris.exceptions import LutrisError
from lutris.gui.dialogs import ErrorDialog
from lutris.util.log import logger
from lutris.util.strings import gtk_safe


class ScriptingError(LutrisError):
    """Custom exception for scripting errors, can be caught by modifying
    excepthook."""

    def __init__(self, message, message_markup=None, faulty_data=None):
        self.faulty_data = faulty_data
        super().__init__(message, message_markup=message_markup)
        logger.error(self.__str__())

    @staticmethod
    def wrap(error: BaseException) -> "ScriptingError":
        if isinstance(error, LutrisError):
            return ScriptingError(error.message, message_markup=error.message_markup)
        else:
            return ScriptingError(str(error))

    def __str__(self):
        if self.faulty_data is None:
            return self.message

        faulty_data = repr(self.faulty_data)
        if not faulty_data:
            return faulty_data

        return self.message + "\n%s" % faulty_data

    def __repr__(self):
        return self.message


class MissingGameDependencyError(LutrisError):
    """Raise when a game requires another game that isn't installed"""

    def __init__(self, *args, message=None, slug=None, **kwargs):
        self.slug = slug

        if not message:
            message = _("This game requires %s.") % slug
        super().__init__(message, *args, **kwargs)


_excepthook = sys.excepthook  # pylint: disable=invalid-name


def error_handler(error_type, value, traceback):
    """Intercept all possible exceptions and raise them as ScriptingErrors"""
    if error_type == ScriptingError:
        message = value.message
        if value.faulty_data:
            message += "\n<b>%s</b>" % gtk_safe(value.faulty_data)
        ErrorDialog(message)
    else:
        _excepthook(error_type, value, traceback)


sys.excepthook = error_handler



import importlib

from lutris import settings
from lutris.util.log import logger

MIGRATION_VERSION = 17  # Never decrease this number

# Replace deprecated migrations with empty lists
MIGRATIONS = [
    [],
    [],
    [],
    [],
    [],
    [],
    [],
    ["mess_to_mame"],
    ["migrate_hidden_ids"],
    ["migrate_steam_appids"],
    [],
    ["retrieve_discord_appids"],
    ["migrate_sortname"],
    ["migrate_hidden_category"],
    ["migrate_ge_proton"],
    ["migrate_banners_back"],
    ["migrate_proton_to_wine_dir"],
]


def get_migration_module(migration_name):
    return importlib.import_module("lutris.migrations.%s" % migration_name)


def migrate():
    current_version = int(settings.read_setting("migration_version") or 0)
    if current_version >= MIGRATION_VERSION:
        return
    for i in range(current_version, MIGRATION_VERSION):
        for migration_name in MIGRATIONS[i]:
            logger.info("Running migration: %s", migration_name)
            migration = get_migration_module(migration_name)
            migration.migrate()

    settings.write_setting("migration_version", MIGRATION_VERSION)



"""Replace the Wine version 'GE-Proton (Latest)' with 'ge-proton'."""

import os

from lutris import settings
from lutris.database.games import get_games_by_runner
from lutris.util.yaml import read_yaml_from_file, write_yaml_to_file


def migrate():
    """Run migration"""
    try:
        config_paths = [os.path.join(settings.CONFIG_DIR, "runners/wine.yml")]

        for db_game in get_games_by_runner("wine"):
            config_filename = db_game.get("configpath")
            config_paths.append(os.path.join(settings.CONFIG_DIR, "games/%s.yml" % config_filename))

        for config_path in config_paths:
            try:
                if os.path.isfile(config_path):
                    config = read_yaml_from_file(config_path)
                    wine = config.get("wine")
                    if wine:
                        version = wine.get("version")
                        if version and version.casefold() == "ge-proton (latest)":
                            wine["version"] = "ge-proton"
                            write_yaml_to_file(config, filepath=config_path)
            except Exception as ex:
                print(f"Failed to convert GE-Proton (Latest) to ge-proton in '{config_path}': {ex}")
    except Exception as ex:
        print(f"Failed to convert GE-Proton (Latest) to ge-proton: {ex}")
        return []
